# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""memoize tests."""

import unittest

from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class CacheTestClass(object):
  """Test cache class."""

  def __init__(self):
    self.called = []

  @memoize.wrap(memoize.FifoInMemory(10))
  def func(self, a, b=2):
    self.called.append((a, b))
    return a + b

  @memoize.wrap(memoize.FifoInMemory(10))
  def none(self, a):
    self.called.append(a)


CALLED = []


def gen_func():
  """Generate function with memoization."""
  del CALLED[:]

  @memoize.wrap(memoize.FifoInMemory(10))
  def func(a, b=2):
    CALLED.append((a, b))
    return a + b

  return func


class WrapTest(unittest.TestCase):
  """Test wrap() decorator on an instance method."""

  def _test_args(self, fn, called):
    """Test args."""
    self.assertEqual(fn(1), 3)
    self.assertListEqual([(1, 2)], called)

    self.assertEqual(fn(1), 3)
    self.assertListEqual([(1, 2)], called)

    self.assertEqual(fn(1, 2), 3)
    self.assertListEqual([(1, 2), (1, 2)], called)

    self.assertEqual(fn(1, 2), 3)
    self.assertListEqual([(1, 2), (1, 2)], called)

  def _test_kwargs(self, fn, called):
    """Test kwargs."""
    self.assertEqual(fn(a=1), 3)
    self.assertListEqual([(1, 2)], called)

    self.assertEqual(fn(a=1), 3)
    self.assertListEqual([(1, 2)], called)

    self.assertEqual(fn(1, b=2), 3)
    self.assertListEqual([(1, 2), (1, 2)], called)

    self.assertEqual(fn(1, b=2), 3)
    self.assertListEqual([(1, 2), (1, 2)], called)

    self.assertEqual(fn(a=1, b=2), 3)
    self.assertListEqual([(1, 2), (1, 2), (1, 2)], called)

    self.assertEqual(fn(a=1, b=2), 3)
    self.assertListEqual([(1, 2), (1, 2), (1, 2)], called)

  def test_args(self):
    """Test args."""
    obj = CacheTestClass()
    self._test_args(obj.func, obj.called)
    self._test_args(gen_func(), CALLED)

  def test_kwargs(self):
    """Test kwargs."""
    obj = CacheTestClass()
    self._test_kwargs(obj.func, obj.called)
    self._test_kwargs(gen_func(), CALLED)

  def test_force_update(self):
    """Test force update."""
    func = gen_func()
    self.assertEqual(func(1), 3)
    self.assertListEqual([(1, 2)], CALLED)

    # pylint: disable=unexpected-keyword-arg
    self.assertEqual(func(1, __memoize_force__=True), 3)
    self.assertListEqual([(1, 2), (1, 2)], CALLED)

  def test_none(self):
    """Test that none means having no key."""
    obj = CacheTestClass()

    self.assertIsNone(obj.none(a=1))
    self.assertListEqual([1], obj.called)

    self.assertIsNone(obj.none(a=1))
    self.assertIsNone(obj.none(a=1))
    self.assertIsNone(obj.none(a=1))

    # Notice that it's called multiple times; this means None is not cached.
    # This actually looks weird. But it's what we want. We might want to fix
    # it later.
    self.assertListEqual([1, 1, 1, 1], obj.called)


class FifoInMemoryTest(unittest.TestCase):
  """Test FifoInMemory."""

  def setUp(self):
    self.cache = memoize.FifoInMemory(5)

  def test_get(self):
    """Test store and get."""

    def fn():
      pass

    key = self.cache.get_key(fn, ('a', 'b'), {'c': 'd'})

    self.assertIsNone(self.cache.get(key))
    self.cache.put(key, 'b')
    self.assertEqual('b', self.cache.get(key))

  def test_hit_limit(self):
    """Test hitting the limit."""
    for i in range(6):
      self.cache.put(i, 'a')

    self.assertIsNone(self.cache.get(0))

    for i in range(1, 6):
      self.assertEqual('a', self.cache.get(i))


class FifoOnDiskTest(fake_filesystem_unittest.TestCase):
  """Test FifoOnDisk."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    environment.set_value('CACHE_DIR', '/tmp/test-cache')
    persistent_cache.initialize()

    self.cache = memoize.FifoOnDisk(5)

  def test_get(self):
    """Test store and get."""

    def fn():
      pass

    key = self.cache.get_key(fn, ('a', 'b'), {'c': 'd'})

    self.assertIsNone(self.cache.get(key))
    self.cache.put(key, 'b')
    self.assertEqual('b', self.cache.get(key))

  def test_hit_limit(self):
    """Test hitting the limit."""
    for i in range(6):
      self.cache.put(i, 'a')

    self.assertIsNone(self.cache.get(0))

    for i in range(1, 6):
      self.assertEqual('a', self.cache.get(i))


class _MockRedis(object):
  """Mock redis client."""

  def __init__(self):
    self._store = {}

  def get(self, key):
    """Get a value."""
    assert isinstance(key, six.string_types)
    return self._store.get(key)

  def set(self, key, value, ex=None):  # pylint: disable=unused-argument
    """Set a value."""
    assert isinstance(key, six.string_types)
    assert isinstance(value, six.string_types)
    self._store[key] = value


class MemcacheTest(unittest.TestCase):
  """Test Memcache."""

  def setUp(self):
    test_helpers.patch(self, [
        'redis.Redis',
        'clusterfuzz._internal.system.environment.is_running_on_app_engine',
    ])
    self.mock.Redis.return_value = _MockRedis()
    self.mock.is_running_on_app_engine.return_value = True

    self.cache = memoize.Memcache(100)

    def fn():
      pass

    self.key = self.cache.get_key(fn, ('a', 'b'), {'c': 'd'})
    self.value = 'b'

  def test_get(self):
    """Test store and get."""
    self.assertIsNone(self.cache.get(self.key))
    self.cache.put(self.key, self.value)
    self.assertEqual(self.value, self.cache.get(self.key))

  def test_noop(self):
    """Test noop on bot."""
    self.mock.is_running_on_app_engine.return_value = False

    self.assertIsNone(self.cache.get(self.key))
    self.cache.put(self.key, self.value)
    self.assertIsNone(self.cache.get(self.key))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for showing crash stats inside the testcase detail page."""

from flask import request

from handlers import base_handler
from libs import crash_stats
from libs import handler
from libs import helpers


def get_result(testcase, end, block, days, group_by):
  """Get slots for crash stats."""
  query = crash_stats.Query()
  query.group_by = group_by
  query.sort_by = 'total_count'
  query.set_time_params(end, days, block)

  query.filter('crash_type', testcase.crash_type)
  query.filter('crash_state', testcase.crash_state)
  query.filter('security_flag', testcase.security_flag)

  _, rows = crash_stats.get(query, crash_stats.Query(), 0, 1)

  if not rows:
    return {'end': end, 'days': days, 'block': block, 'groups': []}

  return rows[0]


class Handler(base_handler.Handler):
  """Serve crash stats data on testcase detail."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_testcase_access
  def post(self, testcase):
    """Server crash stats."""
    end = helpers.cast(request.get('end'), int, "'end' is not an int.")
    days = helpers.cast(request.get('days'), int, "'days' is not an int.")
    group_by = helpers.cast(
        request.get('groupBy'), str, "'groupBy' is not a string.")
    block = helpers.cast(request.get('block'), str, "'block' is not a string.")
    return self.render_json(get_result(testcase, end, block, days, group_by))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Build keyword attributes for every job."""

from google.cloud import ndb

from clusterfuzz._internal.datastore import data_types


def execute(args):
  """Build keywords for jobs."""
  jobs = list(data_types.Job.query())
  if args.non_dry_run:
    ndb.put_multi(jobs)
    print("Done building keywords for jobs.")



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Highlights and generates suppressions for LSAN reports."""

import os
import re

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

# Constants for highlighting.
DIRECT_LEAK_LABEL = 'Direct-leak'
DIRECT_LEAK_REGEX = re.compile(r'^ *Direct leak of')
FIRST_LEAK_DIVIDER = ('%s\nThe following leaks are not necessarily related '
                      'to the first leak.\n\n' % ('=' * 80))
STACK_REGEX = re.compile(r'^ *#[0-9]+\s0x[A-Za-z0-9]+')
STACK_START_REGEX = re.compile(r'^ *#0 ')
BLANK_LINE_REGEX = re.compile(r'^\s*$')

LSAN_TOOL_NAME = 'lsan'
LSAN_SUPPRESSION_LINE = 'leak:{function}\n'
LSAN_HEADER_COMMENT = '# This is a LSAN suppressions file.\n'


def create_empty_local_blacklist():
  """Creates an empty local blacklist."""
  lsan_suppressions_path = get_local_blacklist_file_path()
  with open(lsan_suppressions_path, 'w') as local_blacklist:
    # Insert comment on top to avoid parsing errors on empty file.
    local_blacklist.write(LSAN_HEADER_COMMENT)


def cleanup_global_blacklist():
  """Cleans out closed and deleted testcases from the global blacklist."""
  blacklists_to_delete = []
  global_blacklists = data_types.Blacklist.query(
      data_types.Blacklist.tool_name == LSAN_TOOL_NAME)
  for blacklist in global_blacklists:
    testcase_id = blacklist.testcase_id

    try:
      testcase = data_handler.get_testcase_by_id(testcase_id)
    except errors.InvalidTestcaseError:
      testcase = None

    # Delete entry if testcase is closed, deleted, or unreproducible.
    if not testcase or not testcase.open or testcase.one_time_crasher_flag:
      blacklists_to_delete.append(blacklist.key)

  ndb_utils.delete_multi(blacklists_to_delete)


def copy_global_to_local_blacklist(excluded_testcase=None):
  """Copies contents of global blacklist into local blacklist file, excluding
  a particular testcase (if any)."""
  lsan_suppressions_path = get_local_blacklist_file_path()
  excluded_function_name = (
      get_leak_function_for_blacklist(excluded_testcase)
      if excluded_testcase else None)

  with open(lsan_suppressions_path, 'w') as local_blacklist:
    # Insert comment on top to avoid parsing errors on empty file.
    local_blacklist.write(LSAN_HEADER_COMMENT)

    # Copy global blacklist into local blacklist.
    global_blacklists = data_types.Blacklist.query(
        data_types.Blacklist.tool_name == LSAN_TOOL_NAME)
    blacklisted_functions = []
    for blacklist in global_blacklists:
      if blacklist.function_name in blacklisted_functions:
        continue
      if blacklist.function_name == excluded_function_name:
        continue

      local_blacklist.write(
          LSAN_SUPPRESSION_LINE.format(function=blacklist.function_name))
      blacklisted_functions.append(blacklist.function_name)


def get_leak_function_for_blacklist(testcase):
  """Return leak function to be used for blacklisting."""
  crash_functions = testcase.crash_state.splitlines()
  if not crash_functions:
    return None

  return crash_functions[0]


def get_local_blacklist_file_path():
  """Return the file path to the local blacklist text file."""
  local_blacklist_path = os.path.join(environment.get_suppressions_directory(),
                                      'lsan_suppressions.txt')

  # Create the directory if it does not exists, since we need to write to it.
  blacklist_directory = os.path.dirname(local_blacklist_path)
  if not os.path.exists(blacklist_directory):
    os.makedirs(blacklist_directory)

  return local_blacklist_path


def should_be_blacklisted(testcase):
  """Returns True if testcase is reproducible and not deleted."""
  return (testcase.open and testcase.crash_type == DIRECT_LEAK_LABEL and
          not testcase.one_time_crasher_flag)


def add_crash_to_global_blacklist_if_needed(testcase):
  """Adds relevant function from testcase crash state to global blacklist."""
  testcase_id = testcase.key.id()
  if not should_be_blacklisted(testcase):
    logs.log('Testcase %s is not a reproducible leak, skipping leak blacklist.'
             % testcase_id)
    return False

  function_name = get_leak_function_for_blacklist(testcase)
  if not function_name:
    logs.log_error(
        'Testcase %s has invalid crash state, skipping leak blacklist.' %
        testcase_id)
    return False

  existing_query = data_types.Blacklist.query(
      data_types.Blacklist.function_name == function_name)
  existing_query = existing_query.filter(
      data_types.Blacklist.testcase_id == testcase_id)
  existing_query = existing_query.filter(
      data_types.Blacklist.tool_name == LSAN_TOOL_NAME)

  if existing_query.get():
    logs.log_error('Item already in leak blacklist.')
    return False

  blacklist_item = data_types.Blacklist(
      function_name=function_name,
      testcase_id=testcase_id,
      tool_name=LSAN_TOOL_NAME)
  blacklist_item.put()
  logs.log('Added %s to leak blacklist.' % function_name)

  return blacklist_item


def highlight_first_direct_leak(crash_stacktrace):
  """Highlights the first direct leak in a report.

  Args:
    crash_stacktrace: The crash report.

  Returns:
    new_report: Updated crash report with first direct leak highlighted.
  """
  new_report = []
  processed_first_leak = False
  num_stacks = 0
  highlighted_stack_index = 0
  divider_index = 0

  # Used to prevent highlighting on first indirect leak.
  direct_leak = False
  currently_highlighting = False

  for line in crash_stacktrace.splitlines():
    if DIRECT_LEAK_REGEX.match(line):
      direct_leak = True

    # Marking the end of the highlighted stack with a divider.
    if BLANK_LINE_REGEX.match(line) and currently_highlighting:
      currently_highlighting = False
      processed_first_leak = True

    if STACK_REGEX.match(line):
      if STACK_START_REGEX.match(line):
        num_stacks += 1

        if direct_leak and not processed_first_leak:
          highlighted_stack_index = num_stacks
          currently_highlighting = True

      # If the line is in the first stack, highlight.
      if currently_highlighting:
        line = '<b>%s</b>' % line

    if not processed_first_leak:
      divider_index += 1

    new_report.append(line)

  # If there's only one stack, return original report.
  if num_stacks == 1:
    return crash_stacktrace

  # If there are leaks after the highlighted leak, insert a divider.
  if highlighted_stack_index != num_stacks:
    new_report.insert(divider_index + 1, FIRST_LEAK_DIVIDER)

  return '\n'.join(new_report)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for objects stored in the datastore."""

import re

from google.cloud import ndb
import six

from clusterfuzz._internal.base import json_utils
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import search_tokenizer
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

# Prefix used when a large testcase is stored in the blobstore.
BLOBSTORE_STACK_PREFIX = 'BLOB_KEY='

# List of builtin fuzzers.
BUILTIN_FUZZERS = ['afl', 'libFuzzer']

# Time to look back to find a corpus backup that is marked public.
CORPUS_BACKUP_PUBLIC_LOOKBACK_DAYS = 30

# Marker to indicate end of crash stacktrace. Anything after that is excluded
# from being stored as part of crash stacktrace (e.g. merge content, etc).
CRASH_STACKTRACE_END_MARKER = 'CRASH OUTPUT ENDS HERE'

# Skips using crash state similarity for these types.
CRASH_TYPES_WITH_UNIQUE_STATE = [
    'Missing-library',
    'Out-of-memory',
    'Overwrites-const-input',
    'Timeout',
    # V8 correctness failures use metadata from the fuzz test cases as crash
    # state. This is not suitable for using levenshtein distance for
    # similarity.
    'V8 correctness failure',
]

# Maximum size allowed for an appengine entity type.
# Explicily kept slightly lower than 1 MB.
ENTITY_SIZE_LIMIT = 900000

# Minimum number of unreproducible crashes to see before filing it.
FILE_UNREPRODUCIBLE_TESTCASE_MIN_CRASH_THRESHOLD = 100

# Heartbeat wait interval.
HEARTBEAT_WAIT_INTERVAL = 10 * 60

# FIXME: Move this to configuration.
# List of internal sandboxed data types. This gives a warning on testcase
# uploads on unsandboxed job types.
INTERNAL_SANDBOXED_JOB_TYPES = [
    'linux_asan_chrome_media', 'linux_asan_chrome_mp',
    'linux_asan_chrome_v8_arm', 'mac_asan_chrome', 'windows_asan_chrome'
]

# Time to wait after a report is marked fixed and before filing another similar
# one (hours).
MIN_ELAPSED_TIME_SINCE_FIXED = 2 * 24

# Time to wait for grouping task to finish, before filing the report (hours).
MIN_ELAPSED_TIME_SINCE_REPORT = 4

# Valid name check for fuzzer, job, etc.
NAME_CHECK_REGEX = re.compile(r'^[a-zA-Z0-9_-]+$')

# Regex to match special chars in project name.
SPECIAL_CHARS_REGEX = re.compile('[^a-zA-Z0-9_-]')

# List of supported platforms.
PLATFORMS = [
    'LINUX',
    'ANDROID',
    'CHROMEOS',
    'MAC',
    'WINDOWS',
    'FUCHSIA',
    'ANDROID_KERNEL',
    'ANDROID_AUTO',
]

# Maximum size allowed for an appengine pubsub request.
# Explicily kept slightly lower than 1 MB.
PUBSUB_REQUEST_LIMIT = 900000

# We store at most 3 stacktraces per Testcase entity (original, second, latest).
STACKTRACE_LENGTH_LIMIT = ENTITY_SIZE_LIMIT // 3

# Maximum size allowed for testcase comments.
# 1MiB (maximum Datastore entity size) - ENTITY_SIZE_LIMIT (our limited entity
# size with breathing room), divided by 2 to leave room for other things in the
# entity. This is around 74KB.
TESTCASE_COMMENTS_LENGTH_LIMIT = (1024 * 1024 - ENTITY_SIZE_LIMIT) // 2

# Maximum number of testcase entities to query in one batch.
TESTCASE_ENTITY_QUERY_LIMIT = 256

# Deadlines for testcase filing, closures and deletions (in days).
DUPLICATE_TESTCASE_NO_BUG_DEADLINE = 3
CLOSE_TESTCASE_WITH_CLOSED_BUG_DEADLINE = 14
FILE_CONSISTENT_UNREPRODUCIBLE_TESTCASE_DEADLINE = 14
NOTIFY_CLOSED_BUG_WITH_OPEN_TESTCASE_DEADLINE = 7
UNREPRODUCIBLE_TESTCASE_NO_BUG_DEADLINE = 7
UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE = 14

# Chromium specific issue state tracking labels.
CHROMIUM_ISSUE_RELEASEBLOCK_BETA_LABEL = 'ReleaseBlock-Beta'
# TODO(ochang): Find some way to remove these.
CHROMIUM_ISSUE_PREDATOR_AUTO_CC_LABEL = 'Test-Predator-Auto-CC'
CHROMIUM_ISSUE_PREDATOR_AUTO_COMPONENTS_LABEL = 'Test-Predator-Auto-Components'
CHROMIUM_ISSUE_PREDATOR_AUTO_OWNER_LABEL = 'Test-Predator-Auto-Owner'
CHROMIUM_ISSUE_PREDATOR_WRONG_COMPONENTS_LABEL = (
    'Test-Predator-Wrong-Components')
CHROMIUM_ISSUE_PREDATOR_WRONG_CL_LABEL = 'Test-Predator-Wrong-CLs'

MISSING_VALUE_STRING = '---'

COVERAGE_INFORMATION_DATE_FORMAT = '%Y%m%d'


def clone_entity(e, **extra_args):
  """Clones a DataStore entity and returns the clone."""
  ent_class = e.__class__
  # pylint: disable=protected-access
  props = dict((v._code_name, v.__get__(e, ent_class))
               for v in six.itervalues(ent_class._properties)
               if not isinstance(v, ndb.ComputedProperty))
  props.update(extra_args)
  return ent_class(**props)


class SecuritySeverity(object):
  """Enum for Security Severity."""
  CRITICAL = 0
  HIGH = 1
  MEDIUM = 2
  LOW = 3
  MISSING = 4

  @classmethod
  def is_valid(cls, security_severity):
    """Return bool on whether a severity is valid."""
    return (security_severity in [cls.CRITICAL, cls.HIGH, cls.MEDIUM, cls.LOW])

  @classmethod
  def list(cls):
    """Return the list of severities for a dropdown menu."""
    return [
        {
            'value': cls.CRITICAL,
            'name': 'Critical'
        },
        {
            'value': cls.HIGH,
            'name': 'High',
            'default': True
        },
        {
            'value': cls.MEDIUM,
            'name': 'Medium'
        },
        {
            'value': cls.LOW,
            'name': 'Low'
        },
        {
            'value': cls.MISSING,
            'name': 'Missing'
        },
    ]


# Impact values for security issues.
class SecurityImpact(object):
  EXTENDED_STABLE = 0
  STABLE = 1
  BETA = 2
  HEAD = 3
  NONE = 4
  MISSING = 5


# Archive state enums.
class ArchiveStatus(object):
  NONE = 0
  FUZZED = 1
  MINIMIZED = 2
  ALL = FUZZED | MINIMIZED


# ExternalUserPermission Auto-CC type.
class AutoCCType(object):
  # Don't Auto-CC user.
  NONE = 0
  # Auto-CC user for all issues.
  ALL = 1
  # Auto-CC only for security issues.
  SECURITY = 2


# Type of permission. Used by ExternalUserPermision.
class PermissionEntityKind(object):
  FUZZER = 0
  JOB = 1
  UPLOADER = 2


# Task state string mappings.
class TaskState(object):
  STARTED = 'started'
  WIP = 'in-progress'
  FINISHED = 'finished'
  ERROR = 'errored out'
  NA = ''


# Build state.
class BuildState(object):
  UNMARKED = 0
  GOOD = 1
  BAD = 2


class TestcaseVariantStatus(object):
  PENDING = 0
  REPRODUCIBLE = 1
  FLAKY = 2
  UNREPRODUCIBLE = 3


class Model(ndb.Model):
  """Cache-less NDB model."""
  _use_cache = False
  _use_memcache = False


class Blacklist(Model):
  """Represents global blacklist to track entries for suppressions files."""
  # Function name.
  function_name = ndb.StringProperty()

  # Tool name.
  tool_name = ndb.StringProperty()

  # Testcase ID.
  testcase_id = ndb.IntegerProperty()


class Fuzzer(Model):
  """Represents a fuzzer."""

  VALID_NAME_REGEX = NAME_CHECK_REGEX

  # Last update time.
  timestamp = ndb.DateTimeProperty()

  # Fuzzer Name.
  name = ndb.StringProperty()

  # The name of the archive that the user uploaded.
  filename = ndb.StringProperty()

  # Blobstore key for this fuzzer.
  blobstore_key = ndb.StringProperty()

  # String representation of the file size.
  file_size = ndb.StringProperty()

  # Fuzzer's main executable path, relative to root.
  executable_path = ndb.StringProperty()

  # Revision number of the fuzzer.
  revision = ndb.IntegerProperty()

  # Fuzzer's source (for accountability).
  source = ndb.StringProperty()

  # Testcase timeout.
  timeout = ndb.IntegerProperty()

  # Supported platforms.
  supported_platforms = ndb.StringProperty()

  # Custom script that should be used to launch chrome for this fuzzer.
  launcher_script = ndb.StringProperty()

  # Result from the last fuzzer run showing the number of testcases generated.
  result = ndb.StringProperty()

  # Last result update timestamp.
  result_timestamp = ndb.DateTimeProperty()

  # Console output from last fuzzer run.
  console_output = ndb.TextProperty()

  # Return code from last fuzzer run.
  return_code = ndb.IntegerProperty()

  # Blobstore key for the sample testcase generated by the fuzzer.
  sample_testcase = ndb.StringProperty()

  # Job types for this fuzzer.
  jobs = ndb.StringProperty(repeated=True)

  # Is the fuzzer coming from an external contributor ? Useful for adding
  # reward flags.
  external_contribution = ndb.BooleanProperty(default=False)

  # Max testcases to generate for this fuzzer.
  max_testcases = ndb.IntegerProperty()

  # Does it run un-trusted content ? Examples including running live sites.
  untrusted_content = ndb.BooleanProperty(default=False)

  # Data bundle name.
  data_bundle_name = ndb.StringProperty(default='')

  # Additional environment variables that need to be set for this fuzzer.
  additional_environment_string = ndb.TextProperty()

  # Column specification for stats.
  stats_columns = ndb.TextProperty()

  # Helpful descriptions for the stats_columns. In a yaml format.
  stats_column_descriptions = ndb.TextProperty(indexed=False)

  # Whether this is a builtin fuzzer.
  builtin = ndb.BooleanProperty(indexed=False, default=False)

  # Whether this is a differential fuzzer.
  differential = ndb.BooleanProperty(default=False)

  # If this flag is set, fuzzer generates the testcase in the larger directory
  # on disk |FUZZ_INPUTS_DISK|, rather than smaller tmpfs one (FUZZ_INPUTS).
  has_large_testcases = ndb.BooleanProperty(default=False)


class BuildCrashStatsJobHistory(Model):
  """Represents the record of build_crash_stats run."""
  # End time in hours from epoch, inclusively.
  end_time_in_hours = ndb.IntegerProperty()


class Testcase(Model):
  """Represents a single testcase."""
  # Crash on an invalid read/write.
  crash_type = ndb.StringProperty()

  # Crashing address.
  crash_address = ndb.TextProperty()

  # First x stack frames.
  crash_state = ndb.StringProperty()

  # Complete stacktrace.
  crash_stacktrace = ndb.TextProperty(indexed=False)

  # Last tested crash stacktrace using the latest revision.
  last_tested_crash_stacktrace = ndb.TextProperty(indexed=False)

  # Blobstore keys for various things like original testcase, minimized
  # testcase, etc.
  fuzzed_keys = ndb.TextProperty()
  minimized_keys = ndb.TextProperty()
  minidump_keys = ndb.TextProperty()

  # Tracking issue tracker bug. One bug number per line (future extension).
  bug_information = ndb.StringProperty()

  # Regression range.
  regression = ndb.StringProperty(default='')

  # Revisions where this issue has been fixed.
  fixed = ndb.StringProperty(default='')

  # Is it a security bug ?
  security_flag = ndb.BooleanProperty(default=False)

  # Security severity of the bug.
  security_severity = ndb.IntegerProperty(indexed=False)

  # Did the bug only reproduced once ?
  one_time_crasher_flag = ndb.BooleanProperty(default=False)

  # Any additional comments.
  comments = ndb.TextProperty(default='', indexed=False)

  # Revision that we discovered the crash in.
  crash_revision = ndb.IntegerProperty()

  # The file on the bot that generated the testcase.
  absolute_path = ndb.TextProperty()

  # Minimized argument list.
  minimized_arguments = ndb.TextProperty(default='', indexed=False)

  # Window argument (usually width, height, top, left, etc).
  window_argument = ndb.TextProperty(default='', indexed=False)

  # Type of job associated with this testcase.
  job_type = ndb.StringProperty()

  # Original job queue used for tasks created for this testcase.
  queue = ndb.TextProperty()

  # State representing whether the fuzzed or minimized testcases are archived.
  archive_state = ndb.IntegerProperty(default=0, indexed=False)

  # File name of the original uploaded archive.
  archive_filename = ndb.TextProperty()

  # Is this a binary file?
  binary_flag = ndb.BooleanProperty(default=False, indexed=False)

  # Timestamp.
  timestamp = ndb.DateTimeProperty()

  # Does the testcase crash stack vary b/w crashes ?
  flaky_stack = ndb.BooleanProperty(default=False, indexed=False)

  # Do we need to test this testcase using an HTTP/HTTPS server?
  http_flag = ndb.BooleanProperty(default=False, indexed=False)

  # Name of the fuzzer used to generate this testcase.
  fuzzer_name = ndb.StringProperty()

  # Status of this testcase (pending, processed, unreproducible, etc).
  status = ndb.StringProperty(default='Processed')

  # Id of the testcase that this is marked as a duplicate of.
  duplicate_of = ndb.IntegerProperty(indexed=False)

  # Flag indicating whether or not the testcase has been symbolized.
  symbolized = ndb.BooleanProperty(default=False, indexed=False)

  # Id for this testcase's associated group.
  group_id = ndb.IntegerProperty(default=0)

  # Tracking issue tracker bug for this testcase group.
  group_bug_information = ndb.IntegerProperty(default=0)

  # Fake user interaction sequences like key clicks, mouse movements, etc.
  gestures = ndb.TextProperty(repeated=True)

  # ASAN redzone size in bytes.
  redzone = ndb.IntegerProperty(default=128, indexed=False)

  # Flag indicating if UBSan detection should be disabled. This is needed for
  # cases when ASan and UBSan are bundled in the same build configuration
  # and we need to disable UBSan in some runs to find the potentially more
  # interesting ASan bugs.
  disable_ubsan = ndb.BooleanProperty(default=False)

  # Whether testcase is open.
  open = ndb.BooleanProperty(default=True)

  # Adjusts timeout based on multiplier value.
  timeout_multiplier = ndb.FloatProperty(default=1.0, indexed=False)

  # Additional metadata stored as a JSON object. This should be used for
  # properties that are not commonly accessed and do not need to be indexed.
  additional_metadata = ndb.TextProperty(indexed=False)

  # Boolean attribute indicating if cleanup triage needs to be done.
  triaged = ndb.BooleanProperty(default=False)

  # Project name associated with this test case.
  project_name = ndb.StringProperty()

  # keywords is used for searching.
  keywords = ndb.StringProperty(repeated=True)

  # Whether testcase has a bug (either bug_information or
  # group_bug_information).
  has_bug_flag = ndb.BooleanProperty()

  # Indices for bug_information and group_bug_information.
  bug_indices = ndb.StringProperty(repeated=True)

  # Overridden fuzzer name because actual fuzzer name can be different in many
  # scenarios (libfuzzer, afl, etc).
  overridden_fuzzer_name = ndb.StringProperty()

  # Platform (e.g. windows, linux, android).
  platform = ndb.StringProperty()

  # Platform id (e.g. windows, linux, android:hammerhead:l).
  # For Android, includes device type and underlying OS version.
  platform_id = ndb.StringProperty()

  # Impact indices for searching.
  impact_indices = ndb.StringProperty(repeated=True)

  # Whether or not a testcase is a duplicate of other testcase.
  is_a_duplicate_flag = ndb.BooleanProperty()

  # Whether or not a testcase is the leader of its group.
  # If the testcase is not in a group, it's the leader of a group of 1.
  # The default is false because we prefer not to show crashes until we are
  # sure. And group_task will correctly set the value within 30 minutes.
  is_leader = ndb.BooleanProperty(default=False)

  # Fuzzer name indices
  fuzzer_name_indices = ndb.StringProperty(repeated=True)

  # The impacted version indices (including beta, stable and extended_stable).
  impact_version_indices = ndb.StringProperty(repeated=True)

  # The impacted extended stable version.
  impact_extended_stable_version = ndb.StringProperty()

  # The impacted extended stable version indices.
  impact_extended_stable_version_indices = ndb.StringProperty(repeated=True)

  # The impacted extended stable version is merely probable (not definite).
  # See the comment on impact_stable_version_likely.
  impact_extended_stable_version_likely = ndb.BooleanProperty()

  # The impacted stable version.
  impact_stable_version = ndb.StringProperty()

  # The impacted stable version indices.
  impact_stable_version_indices = ndb.StringProperty(repeated=True)

  # The impacted stable version is merely probable (not definite). Because
  # for a non-asan build, we don't have a stable/beta build. Therefore, we
  # make an intelligent guess on the version.
  impact_stable_version_likely = ndb.BooleanProperty()

  # The impacted beta version.
  impact_beta_version = ndb.StringProperty()

  # The impacted beta version indices.
  impact_beta_version_indices = ndb.StringProperty(repeated=True)

  # The impacted beta version is merely probable (not definite). See the
  # comment on impact_stable_version_likely.
  impact_beta_version_likely = ndb.BooleanProperty()

  # The impacted 'head' version
  impact_head_version = ndb.StringProperty()

  # The impacted head version indices.
  impact_head_version_indices = ndb.StringProperty(repeated=True)

  # The impacted head version is likely.
  impact_head_version_likely = ndb.BooleanProperty()

  # Whether or not impact task has been run on this testcase.
  is_impact_set_flag = ndb.BooleanProperty()

  # Uploader email address.
  uploader_email = ndb.StringProperty()

  def has_blame(self):
    return self.project_name == 'chromium'

  def has_impacts(self):
    return self.project_name == 'chromium' and not self.one_time_crasher_flag

  def impacts_production(self):
    return (bool(self.impact_extended_stable_version) or
            bool(self.impact_stable_version) or bool(self.impact_beta_version))

  def is_status_unreproducible(self):
    return self.status and self.status.startswith('Unreproducible')

  def is_crash(self):
    return bool(self.crash_state)

  def populate_indices(self):
    """Populate keywords for fast test case list searching."""
    self.keywords = list(
        search_tokenizer.tokenize(self.crash_state)
        | search_tokenizer.tokenize(self.crash_type)
        | search_tokenizer.tokenize(self.fuzzer_name)
        | search_tokenizer.tokenize(self.overridden_fuzzer_name)
        | search_tokenizer.tokenize(self.job_type)
        | search_tokenizer.tokenize(self.platform_id))

    self.bug_indices = search_tokenizer.tokenize_bug_information(self)
    self.has_bug_flag = bool(self.bug_indices)
    self.is_a_duplicate_flag = bool(self.duplicate_of)
    fuzzer_name_indices = list(
        set([self.fuzzer_name, self.overridden_fuzzer_name]))
    self.fuzzer_name_indices = [f for f in fuzzer_name_indices if f]

    # If the impact task hasn't been run (aka is_impact_set_flag=False) OR
    # if impact isn't applicable (aka has_impacts() is False), we wipe all
    # the impact fields' indices.
    if self.has_impacts() and self.is_impact_set_flag:
      self.impact_extended_stable_version_indices = (
          search_tokenizer.tokenize_impact_version(
              self.impact_extended_stable_version))
      self.impact_stable_version_indices = (
          search_tokenizer.tokenize_impact_version(self.impact_stable_version))
      self.impact_beta_version_indices = (
          search_tokenizer.tokenize_impact_version(self.impact_beta_version))
      self.impact_head_version_indices = (
          search_tokenizer.tokenize_impact_version(self.impact_head_version))
      self.impact_version_indices = list(
          set(self.impact_extended_stable_version_indices +
              self.impact_stable_version_indices +
              self.impact_head_version_indices +
              self.impact_beta_version_indices))
      if self.impact_extended_stable_version:
        self.impact_version_indices.append('extended_stable')
      if self.impact_beta_version:
        self.impact_version_indices.append('beta')
      if self.impact_stable_version:
        self.impact_version_indices.append('stable')
      if not self.impacts_production():
        self.impact_version_indices.append('head')
    else:
      self.impact_version_indices = []
      self.impact_extended_stable_version_indices = []
      self.impact_stable_version_indices = []
      self.impact_beta_version_indices = []

  def _pre_put_hook(self):
    self.populate_indices()

  def _post_put_hook(self, _):
    if not self.key:
      # Failed put. An exception will be thrown automatically afterwards.
      return

    logs.log('Updated testcase %d (bug %s).' % (self.key.id(),
                                                self.bug_information or '-'))

  def set_impacts_as_na(self):
    self.impact_stable_version = self.impact_beta_version = None
    self.impact_extended_stable_version = None
    self.impact_stable_version_likely = self.impact_beta_version_likely = False
    self.impact_extended_stable_version_likely = False
    self.is_impact_set_flag = False

  def _ensure_metadata_is_cached(self):
    """Ensure that the metadata for this has been cached."""
    if hasattr(self, 'metadata_cache'):
      return

    try:
      cache = json_utils.loads(self.additional_metadata)
    except (TypeError, ValueError):
      cache = {}

    setattr(self, 'metadata_cache', cache)

  def get_metadata(self, key=None, default=None):
    """Get metadata for a test case. Slow on first access."""
    self._ensure_metadata_is_cached()

    # If no key is specified, return all metadata.
    if not key:
      return self.metadata_cache

    try:
      return self.metadata_cache[key]
    except KeyError:
      return default

  def set_metadata(self, key, value, update_testcase=True):
    """Set metadata for a test case."""
    self._ensure_metadata_is_cached()
    self.metadata_cache[key] = value

    self.additional_metadata = json_utils.dumps(self.metadata_cache)
    if update_testcase:
      self.put()

  def delete_metadata(self, key, update_testcase=True):
    """Remove metadata key for a test case."""
    self._ensure_metadata_is_cached()

    # Make sure that the key exists in cache. If not, no work to do here.
    if key not in self.metadata_cache:
      return

    del self.metadata_cache[key]
    self.additional_metadata = json_utils.dumps(self.metadata_cache)
    if update_testcase:
      self.put()

  def actual_fuzzer_name(self):
    """Actual fuzzer name, uses one from overridden attribute if available."""
    return self.overridden_fuzzer_name or self.fuzzer_name

  def get_fuzz_target(self):
    """Get the associated FuzzTarget entity for this test case."""
    name = self.actual_fuzzer_name()
    if not name:
      return None

    target = ndb.Key(FuzzTarget, name).get()
    if environment.get_value('ORIGINAL_JOB_NAME'):
      # Overridden engine (e.g. for minimization).
      target.engine = environment.get_engine_for_job()

    return target


class TestcaseGroup(Model):
  """Group for a set of testcases."""


class DataBundle(Model):
  """Represents a data bundle associated with a fuzzer."""

  VALID_NAME_REGEX = NAME_CHECK_REGEX

  # The data bundle's name (important for identifying shared bundles).
  name = ndb.StringProperty()

  # Name of cloud storage bucket on GCS.
  bucket_name = ndb.StringProperty()

  # Data bundle's source (for accountability).
  # TODO(ochang): Remove.
  source = ndb.StringProperty()

  # If data bundle can be unpacked locally or needs nfs.
  is_local = ndb.BooleanProperty(default=True)

  # Creation timestamp.
  timestamp = ndb.DateTimeProperty()

  # Whether or not bundle should be synced to worker instead.
  # Fuzzer scripts are usually run on trusted hosts, so data bundles are synced
  # there. In libFuzzer's case, we want the bundle to be on the same machine as
  # where the libFuzzer binary will run (untrusted).
  sync_to_worker = ndb.BooleanProperty(default=False)


class Config(Model):
  """Configuration."""
  previous_hash = ndb.StringProperty(default='')

  # Project's url.
  url = ndb.StringProperty(default='')

  # Issue tracker client authentication parameters.
  client_credentials = ndb.TextProperty(default='')

  # Jira url and credentials
  jira_url = ndb.StringProperty(default='')
  jira_credentials = ndb.TextProperty(default='')

  # Build apiary authentication parameters.
  build_apiary_service_account_private_key = ndb.TextProperty(default='')

  # Google test account for login, gms testing, etc.
  test_account_email = ndb.StringProperty(default='')
  test_account_password = ndb.StringProperty(default='')

  # Privileged users.
  privileged_users = ndb.TextProperty(default='')

  # Blacklisted users.
  blacklisted_users = ndb.TextProperty(default='')

  # Admin contact string.
  contact_string = ndb.StringProperty(default='')

  # Component to repository mappings.
  component_repository_mappings = ndb.TextProperty(default='')

  # URL for help page for reproducing issues.
  reproduction_help_url = ndb.StringProperty(default='')

  # Documentation url.
  documentation_url = ndb.StringProperty(default='')

  # Bug report url.
  bug_report_url = ndb.StringProperty(default='')

  # Platforms that coverage is supported for.
  platform_group_mappings = ndb.TextProperty(default='')

  # More relaxed restrictions: allow CC'ed users and reporters of issues to view
  # testcase details.
  relax_testcase_restrictions = ndb.BooleanProperty(default=False)

  # More relaxed restrictions: allow domain users to access both security and
  # functional bugs.
  relax_security_bug_restrictions = ndb.BooleanProperty(default=False)

  # Coverage reports bucket.
  coverage_reports_bucket = ndb.StringProperty(default='')

  # For GitHub API.
  github_credentials = ndb.StringProperty(default='')

  # OAuth2 client id for the reproduce tool.
  reproduce_tool_client_id = ndb.StringProperty(default='')

  # OAuth2 client secret for the reproduce tool.
  reproduce_tool_client_secret = ndb.StringProperty(default='')

  # Pub/Sub topics for the Predator service.
  predator_crash_topic = ndb.StringProperty(default='')
  predator_result_topic = ndb.StringProperty(default='')

  # Wifi connection information.
  wifi_ssid = ndb.StringProperty(default='')
  wifi_password = ndb.StringProperty(default='')

  # SendGrid config.
  sendgrid_api_key = ndb.StringProperty(default='')
  sendgrid_sender = ndb.StringProperty(default='')


class TestcaseUploadMetadata(Model):
  """Metadata associated with a user uploaded test case."""
  # Timestamp.
  timestamp = ndb.DateTimeProperty()

  # Testcase filename.
  filename = ndb.StringProperty()

  # Current status of the testcase.
  status = ndb.StringProperty()

  # Uploader email address.
  uploader_email = ndb.StringProperty()

  # Name of the bot that ran analyze on this testcase.
  bot_name = ndb.StringProperty()

  # Id of the associated testcase.
  testcase_id = ndb.IntegerProperty()

  # Id of the testcase that this is marked as a duplicate of.
  duplicate_of = ndb.IntegerProperty()

  # Blobstore key for the testcase associated with this object.
  blobstore_key = ndb.StringProperty()

  # Testcase timeout.
  timeout = ndb.IntegerProperty()

  # Is this a single testcase bundled in an archive?
  bundled = ndb.BooleanProperty()

  # Path to the file in the archive.
  path_in_archive = ndb.TextProperty()

  # Original blobstore key for this object (used for archives).
  original_blobstore_key = ndb.StringProperty()

  # Security flag.
  security_flag = ndb.BooleanProperty(default=False)

  # Number of retries for this testcase.
  retries = ndb.IntegerProperty()

  # Flag to indicate where bug title should be updated or not.
  bug_summary_update_flag = ndb.BooleanProperty()

  # Flag to indicate if we are running in quiet mode (e.g. bug updates).
  quiet_flag = ndb.BooleanProperty()

  # Additional testcase metadata dict stored as a string.
  additional_metadata_string = ndb.TextProperty(indexed=False)

  # Specified issue id.
  bug_information = ndb.StringProperty()


class JobTemplate(Model):
  # Job template name.
  name = ndb.StringProperty()

  # Environment string.
  environment_string = ndb.TextProperty()


class Job(Model):
  """Definition of a job type used by the bots."""

  VALID_NAME_REGEX = NAME_CHECK_REGEX

  # Job type name.
  name = ndb.StringProperty()

  # Job environment string.
  environment_string = ndb.TextProperty()

  # The platform that this job can run on.
  platform = ndb.StringProperty()

  # Blobstore key of the custom binary for this job.
  custom_binary_key = ndb.StringProperty()

  # Filename for the custom binary.
  custom_binary_filename = ndb.StringProperty()

  # Revision of the custom binary.
  custom_binary_revision = ndb.IntegerProperty()

  # Description of the job.
  description = ndb.TextProperty()

  # Template to use, if any.
  templates = ndb.StringProperty(repeated=True)

  # Project name.
  project = ndb.StringProperty()

  # Keywords is used for searching.
  keywords = ndb.StringProperty(repeated=True)

  # If this is set, this Job is for an external reproduction infrastructure. The
  # value here is the topic used for reproduction requests.
  external_reproduction_topic = ndb.StringProperty()

  # If this is set, this Job is for an external reproduction infrastructure. The
  # value here is the subscription used for receiving reproduction updates.
  external_updates_subscription = ndb.StringProperty()

  def is_external(self):
    """Whether this job is external."""
    return (bool(self.external_reproduction_topic) or
            bool(self.external_updates_subscription))

  def get_environment(self):
    """Get the environment as a dict for this job, including any environment
    variables in its template."""
    if not self.templates:
      return environment.parse_environment_definition(self.environment_string)

    job_environment = {}
    for template_name in self.templates:
      template = JobTemplate.query(JobTemplate.name == template_name).get()
      if not template:
        continue

      template_environment = environment.parse_environment_definition(
          template.environment_string)

      job_environment.update(template_environment)

    environment_overrides = environment.parse_environment_definition(
        self.environment_string)

    job_environment.update(environment_overrides)
    return job_environment

  def get_environment_string(self):
    """Get the environment string for this job, including any environment
    variables in its template. Avoid using this if possible."""
    environment_string = ''
    job_environment = self.get_environment()
    for key, value in six.iteritems(job_environment):
      environment_string += '%s = %s\n' % (key, value)

    return environment_string

  def populate_indices(self):
    """Populate keywords for fast job searching."""
    self.keywords = list(
        search_tokenizer.tokenize(self.name)
        | search_tokenizer.tokenize(self.project))

  def _pre_put_hook(self):
    """Pre-put hook."""
    self.project = self.get_environment().get('PROJECT_NAME',
                                              utils.default_project_name())
    self.populate_indices()


class CSRFToken(Model):
  """Token used to prevent CSRF attacks."""
  # Value of this token.
  value = ndb.StringProperty()

  # Expiration time for this token.
  expiration_time = ndb.DateTimeProperty()

  # User that this token is associated with.
  user_email = ndb.StringProperty()


class Heartbeat(Model):
  """Bot health metadata."""
  # Name of the bot.
  bot_name = ndb.StringProperty()

  # Time of the last heartbeat.
  last_beat_time = ndb.DateTimeProperty()

  # Task payload containing information on current task execution.
  task_payload = ndb.StringProperty()

  # Expected end time for task.
  task_end_time = ndb.DateTimeProperty()

  # Source version (for accountability).
  source_version = ndb.StringProperty()

  # Platform id (esp important for Android platform for OS version).
  platform_id = ndb.StringProperty()

  # Keywords is used for searching.
  keywords = ndb.StringProperty(repeated=True)

  def populate_indices(self):
    """Populate keywords for fast job searching."""
    self.keywords = list(
        search_tokenizer.tokenize(self.bot_name)
        | search_tokenizer.tokenize(self.task_payload))

  def _pre_put_hook(self):
    """Pre-put hook."""
    self.populate_indices()


class Notification(Model):
  """Tracks whether or not an email has been sent to a user for a test case."""
  # Testcase id associated with this notification.
  testcase_id = ndb.IntegerProperty()

  # User that this notification was sent to.
  user_email = ndb.StringProperty()


class BundledArchiveMetadata(Model):
  """Metadata needed for multiple test cases uploaded in an archive."""
  # Blobstore key of the archive.
  blobstore_key = ndb.StringProperty()

  # Timeout in seconds for each testcase in the bundle.
  timeout = ndb.IntegerProperty()

  # Job queue for the analyze tasks created for this bundle.
  job_queue = ndb.StringProperty()

  # Job type that should be used for all testcases in this bundle.
  job_type = ndb.StringProperty()

  # Flag indicating whether or not these testcases need http.
  http_flag = ndb.BooleanProperty()

  # File name of the uploaded archive.
  archive_filename = ndb.StringProperty()

  # Email address of the uploader of the archive.
  uploader_email = ndb.StringProperty()

  # Fake user interaction sequences like key clicks, mouse movements, etc.
  gestures = ndb.StringProperty(repeated=True)

  # Optional. Revision that we discovered the crash in.
  crash_revision = ndb.IntegerProperty()

  # Optional. Additional arguments.
  additional_arguments = ndb.StringProperty()

  # Optional. Bug information.
  bug_information = ndb.StringProperty()

  # Optional. Platform id, e.g. android:shamu.
  platform_id = ndb.StringProperty()

  # Optional. App launch command. e.g. shell am start ...
  app_launch_command = ndb.StringProperty()

  # Fuzzer name.
  fuzzer_name = ndb.StringProperty()

  # Overridden fuzzer name because actual fuzzer name can be different in many
  # scenarios (libfuzzer, afl, etc).
  overridden_fuzzer_name = ndb.StringProperty()

  # Binary name for fuzz target (only applicable to libFuzzer, AFL).
  fuzzer_binary_name = ndb.StringProperty()


class TaskStatus(Model):
  """Information about task status."""
  # Bot name.
  bot_name = ndb.StringProperty()

  # Status.
  status = ndb.StringProperty()

  # Time of creation or last update time.
  time = ndb.DateTimeProperty()


class BuildMetadata(Model):
  """Metadata associated with a particular archived build."""
  # Job type that this build belongs to.
  job_type = ndb.StringProperty()

  # Revision of the build.
  revision = ndb.IntegerProperty()

  # Good build or bad build.
  bad_build = ndb.BooleanProperty(default=False)

  # Stdout and stderr.
  console_output = ndb.TextProperty()

  # Bot name.
  bot_name = ndb.StringProperty()

  # Symbol data.
  symbols = ndb.StringProperty()

  # Creation timestamp.
  timestamp = ndb.DateTimeProperty()


class ReportMetadata(Model):
  """Metadata associated with a crash report."""
  # Job type from testcase.
  job_type = ndb.StringProperty()

  # Revision of build from report.
  crash_revision = ndb.IntegerProperty(default=-1)

  # Has this report been successfully uploaded?
  is_uploaded = ndb.BooleanProperty(default=False)

  # Product.
  product = ndb.StringProperty(default='')

  # Version.
  version = ndb.TextProperty(default='')

  # Key to minidump previously written to blobstore.
  minidump_key = ndb.TextProperty(default='')

  # Processed crash bytes.
  serialized_crash_stack_frames = ndb.BlobProperty(default='', indexed=False)

  # Id of the associated testcase.
  testcase_id = ndb.StringProperty(default='')

  # Id of the associated bot.
  bot_id = ndb.TextProperty(default='')

  # Optional upload params, stored as a JSON object.
  optional_params = ndb.TextProperty(indexed=False)

  # Report id from crash/.
  crash_report_id = ndb.StringProperty()


class Lock(Model):
  """Lock entity."""
  # Expiration time for the lock.
  expiration_time = ndb.DateTimeProperty()

  # The bot name denoting the holder of the lock.
  holder = ndb.StringProperty()


class FuzzTarget(Model):
  """Fuzz target."""
  # The engine this target is a child of.
  engine = ndb.StringProperty()

  # Project name.
  project = ndb.StringProperty()

  # Binary name.
  binary = ndb.StringProperty()

  def _pre_put_hook(self):
    """Pre-put hook."""
    self.key = ndb.Key(FuzzTarget, self.fully_qualified_name())

  def fully_qualified_name(self):
    """Get the fully qualified name for this fuzz target."""
    return fuzz_target_fully_qualified_name(self.engine, self.project,
                                            self.binary)

  def project_qualified_name(self):
    """Get the name qualified by project."""
    return fuzz_target_project_qualified_name(self.project, self.binary)


def fuzz_target_fully_qualified_name(engine, project, binary):
  """Get a fuzz target's fully qualified name."""
  return engine + '_' + fuzz_target_project_qualified_name(project, binary)


def normalized_name(name):
  """Return normalized name with special chars like slash, colon, etc normalized
  to hyphen(-). This is important as otherwise these chars break local and cloud
  storage paths."""
  return SPECIAL_CHARS_REGEX.sub('-', name).strip('-')


def fuzz_target_project_qualified_name(project, binary):
  """Get a fuzz target's project qualified name."""
  binary = normalized_name(binary)
  if not project:
    return binary

  if project == utils.default_project_name():
    # Don't prefix with project name if it's the default project.
    return binary

  normalized_project_prefix = normalized_name(project) + '_'
  if binary.startswith(normalized_project_prefix):
    return binary

  return normalized_project_prefix + binary


class FuzzTargetsCount(Model):
  """Fuzz targets count for every job. Key IDs are the job name."""
  count = ndb.IntegerProperty(indexed=False)


class FuzzTargetJob(Model):
  """Mapping between fuzz target and jobs with additional metadata for
  selection."""
  # Fully qualified fuzz target name.
  fuzz_target_name = ndb.StringProperty()

  # Job this target ran as.
  job = ndb.StringProperty()

  # Engine this ran as.
  engine = ndb.StringProperty()

  # Relative frequency with which to select this fuzzer.
  weight = ndb.FloatProperty(default=1.0)

  # Approximate last time this target was run.
  last_run = ndb.DateTimeProperty()

  def _pre_put_hook(self):
    """Pre-put hook."""
    self.key = ndb.Key(FuzzTargetJob,
                       fuzz_target_job_key(self.fuzz_target_name, self.job))


class FuzzStrategyProbability(Model):
  """Mapping between fuzz strategies and probabilities with which they
  should be selected."""

  strategy_name = ndb.StringProperty()
  probability = ndb.FloatProperty()
  engine = ndb.StringProperty()


def fuzz_target_job_key(fuzz_target_name, job):
  """Return the key for FuzzTargetJob."""
  return '{}/{}'.format(fuzz_target_name, job)


class ExternalUserPermission(Model):
  """Permissions for external users."""
  # Email user is authenticated as.
  email = ndb.StringProperty()

  # Type of |entity_name|. Can be one of the values of PermissionEntityKind.
  entity_kind = ndb.IntegerProperty()

  # Name of the entity that user is allowed to view.
  entity_name = ndb.StringProperty()

  # Whether or not |allowed_name| is a prefix.
  is_prefix = ndb.BooleanProperty(default=False)

  # Auto CC type.
  auto_cc = ndb.IntegerProperty()


class FiledBug(Model):
  """Metadata information for issues that were filed automatically."""
  # Timestamp when the issue was filed.
  timestamp = ndb.DateTimeProperty()

  # ID of the test case that is associated with the filed issue.
  testcase_id = ndb.IntegerProperty()

  # Tracking issue tracker bug for this testcase.
  bug_information = ndb.IntegerProperty(default=0)

  # Group ID associated with this issue.
  group_id = ndb.IntegerProperty()

  # Crash type for easy reference.
  crash_type = ndb.StringProperty()

  # Crash state for easy reference.
  crash_state = ndb.StringProperty()

  # Is it a security bug?
  security_flag = ndb.BooleanProperty()

  # Platform id.
  platform_id = ndb.StringProperty()


class CoverageInformation(Model):
  """Coverage info."""
  date = ndb.DateProperty(auto_now_add=True)
  fuzzer = ndb.StringProperty()

  # Function coverage information.
  functions_covered = ndb.IntegerProperty()
  functions_total = ndb.IntegerProperty()

  # Edge coverage information.
  edges_covered = ndb.IntegerProperty()
  edges_total = ndb.IntegerProperty()

  # Corpus size information.
  corpus_size_units = ndb.IntegerProperty()
  corpus_size_bytes = ndb.IntegerProperty()
  corpus_location = ndb.StringProperty()

  # Corpus backup information.
  corpus_backup_location = ndb.StringProperty()

  # Quarantine size information.
  quarantine_size_units = ndb.IntegerProperty()
  quarantine_size_bytes = ndb.IntegerProperty()
  quarantine_location = ndb.StringProperty()

  # Link to the HTML report.
  html_report_url = ndb.StringProperty()

  def _pre_put_hook(self):
    """Pre-put hook."""
    self.key = ndb.Key(CoverageInformation,
                       coverage_information_key(self.fuzzer, self.date))


class CorpusTag(Model):
  """Corpus Tags for sharing corpora between fuzz targets."""
  tag = ndb.StringProperty()
  fully_qualified_fuzz_target_name = ndb.StringProperty()


def coverage_information_date_to_string(date):
  """Returns string representation of the date in a format used for coverage."""
  return date.strftime(COVERAGE_INFORMATION_DATE_FORMAT)


def coverage_information_key(project_qualified_fuzzer_name, date):
  """Constructs an ndb key for CoverageInformation entity."""
  date_string = coverage_information_date_to_string(date)
  return project_qualified_fuzzer_name + '-' + date_string


class Trial(Model):
  """Trials for specific binaries."""
  # App name that this trial is applied to. E.g. "d8" or "chrome".
  app_name = ndb.StringProperty()

  # Chance to select this set of arguments. Zero to one.
  probability = ndb.FloatProperty()

  # Additional arguments to apply if selected.
  app_args = ndb.TextProperty()


# TODO(ochang): Make this generic.
class OssFuzzProject(Model):
  """Represents a project that has been set up for OSS-Fuzz."""
  # Name of the project.
  name = ndb.StringProperty()

  # Whether or not the project should run on high end hosts.
  high_end = ndb.BooleanProperty(default=False)

  # Weight for CPU distribution. This is set by admins.
  cpu_weight = ndb.FloatProperty(default=1.0)

  # The disk size to use (overrides the default).
  disk_size_gb = ndb.IntegerProperty()

  # Service account for this project.
  service_account = ndb.StringProperty()

  # CCs for the project.
  ccs = ndb.StringProperty(repeated=True)


class OssFuzzProjectInfo(Model):
  """Set up information for a project (cpu allocation, instance groups, service
  accounts)."""

  class ClusterInfo(Model):
    """Cpu allocation information for a project in a zone."""
    # The cluster for the CPU allocation.
    cluster = ndb.StringProperty()

    # The number of allocated CPUs in this cluster.
    cpu_count = ndb.IntegerProperty(default=0)

    # The GCE zone for this cluster.
    gce_zone = ndb.StringProperty()

  # Name of the project.
  name = ndb.StringProperty()

  # Information about CPUs in each cluster.
  clusters = ndb.StructuredProperty(ClusterInfo, repeated=True)

  def get_cluster_info(self, name):
    return next((info for info in self.clusters if info.cluster == name), None)


class HostWorkerAssignment(Model):
  """Host worker assignment information."""
  # The host instance name.
  host_name = ndb.StringProperty()

  # The instance number (0 to WORKERS_PER_HOST - 1).
  instance_num = ndb.IntegerProperty()

  # The worker instance name.
  worker_name = ndb.StringProperty()

  # The project name.
  project_name = ndb.StringProperty()


class WorkerTlsCert(Model):
  """TLS certs for untrusted workers."""
  # The name of the project.
  project_name = ndb.StringProperty()

  # The contents of the TLS cert.
  cert_contents = ndb.BlobProperty()

  # The contents of the private key.
  key_contents = ndb.BlobProperty()


class FuzzerJob(Model):
  """Mapping between a fuzzer and job with additional metadata for selection."""
  fuzzer = ndb.StringProperty()
  job = ndb.StringProperty()
  platform = ndb.StringProperty()
  weight = ndb.FloatProperty(default=1.0)
  multiplier = ndb.FloatProperty(default=1.0)

  @property
  def actual_weight(self):
    """Get the actual weight for this job."""
    return self.weight * self.multiplier


class FuzzerJobs(Model):
  """(Batched) mappings between a fuzzer and jobs with additional metadata for
  selection."""
  platform = ndb.StringProperty()
  fuzzer_jobs = ndb.LocalStructuredProperty(FuzzerJob, repeated=True)


class OssFuzzBuildFailure(Model):
  """Represents build failure."""
  # Project name.
  project_name = ndb.StringProperty()

  # The monorail issue ID for the failure.
  issue_id = ndb.StringProperty()

  # The last timestamp of the build.
  last_checked_timestamp = ndb.DateTimeProperty()

  # Number of consecutive failures.
  consecutive_failures = ndb.IntegerProperty(default=0)

  # Build type (fuzzing, coverage, etc).
  build_type = ndb.StringProperty()


class Admin(Model):
  """Records an admin user."""
  email = ndb.StringProperty()


class TestcaseVariant(Model):
  """Represent a testcase variant on another job (another platform / sanitizer
  / config)."""
  # Testcase ID of the testcase for which the variant is being evaluated.
  testcase_id = ndb.IntegerProperty()

  # Status of the testcase variant (pending, reproducible, unreproducible, etc).
  status = ndb.IntegerProperty(default=0)

  # Job type for the testcase variant.
  job_type = ndb.StringProperty()

  # Revision that the testcase variant was tried against.
  revision = ndb.IntegerProperty()

  # Crash type.
  crash_type = ndb.StringProperty()

  # Crash state.
  crash_state = ndb.StringProperty()

  # Bool to indicate if it is a security bug?
  security_flag = ndb.BooleanProperty()

  # Bool to indicate if crash is similar to original testcase.
  is_similar = ndb.BooleanProperty()

  # Similar testcase reproducer key (optional). This is set in case we notice a
  # similar crash on another platform.
  reproducer_key = ndb.StringProperty()

  # Platform (e.g. windows, linux, android).
  platform = ndb.StringProperty()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for redoing."""

from flask import request

from clusterfuzz._internal.base import tasks
from handlers import base_handler
from handlers.testcase_detail import show
from libs import handler
from libs import helpers


class Handler(base_handler.Handler):
  """Handler that redo tasks."""

  @staticmethod
  def redo(testcase, testcase_tasks, user_email):
    """Redo tasks."""
    try:
      tasks.redo_testcase(testcase, testcase_tasks, user_email)
    except tasks.InvalidRedoTask as error:
      raise helpers.EarlyExitException(str(error), 400)

    helpers.log('Redo testcase %d: %s' % (testcase.key.id(), testcase_tasks),
                helpers.MODIFY_OPERATION)

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_testcase_access
  def post(self, testcase):
    """Queue redo tasks."""
    testcase_tasks = request.get('tasks')
    user_email = helpers.get_user_email()

    self.redo(testcase, testcase_tasks, user_email)
    return self.render_json(show.get_testcase_detail(testcase))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Grouper for grouping similar looking testcases."""

import six

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.crash_analysis.crash_comparer import CrashComparer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from libs.issue_management import issue_tracker_utils

from . import group_leader

FORWARDED_ATTRIBUTES = ('crash_state', 'crash_type', 'group_id',
                        'one_time_crasher_flag', 'project_name',
                        'security_flag', 'timestamp')

GROUP_MAX_TESTCASE_LIMIT = 25


class TestcaseAttributes(object):
  """Testcase attributes used for grouping."""

  __slots__ = ('id', 'is_leader', 'issue_id') + FORWARDED_ATTRIBUTES

  def __init__(self, testcase_id):
    self.id = testcase_id
    self.is_leader = True
    self.issue_id = None


def combine_testcases_into_group(testcase_1, testcase_2, testcase_map):
  """Combine two testcases into a group."""
  logs.log(
      'Grouping testcase 1 '
      '(crash_type=%s, crash_state=%s, security_flag=%s, group=%s) '
      'and testcase 2 '
      '(crash_type=%s, crash_state=%s, security_flag=%s, group=%s).' %
      (testcase_1.crash_type, testcase_1.crash_state, testcase_1.security_flag,
       testcase_1.group_id, testcase_2.crash_type, testcase_2.crash_state,
       testcase_2.security_flag, testcase_2.group_id))

  # If none of the two testcases have a group id, just assign a new group id to
  # both.
  if not testcase_1.group_id and not testcase_2.group_id:
    new_group_id = _get_new_group_id()
    testcase_1.group_id = new_group_id
    testcase_2.group_id = new_group_id
    return

  # If one of the testcase has a group id, then assign the other to reuse that
  # group id.
  if testcase_1.group_id and not testcase_2.group_id:
    testcase_2.group_id = testcase_1.group_id
    return
  if testcase_2.group_id and not testcase_1.group_id:
    testcase_1.group_id = testcase_2.group_id
    return

  # If both the testcase have their own groups, then just merge the two groups
  # together and reuse one of their group ids.
  group_id_to_reuse = testcase_1.group_id
  group_id_to_move = testcase_2.group_id
  for testcase in six.itervalues(testcase_map):
    if testcase.group_id == group_id_to_move:
      testcase.group_id = group_id_to_reuse


def _get_new_group_id():
  """Get a new group id for testcase grouping."""
  new_group = data_types.TestcaseGroup()
  new_group.put()
  return new_group.key.id()


def _group_testcases_with_same_issues(testcase_map):
  """Group testcases that are associated with same underlying issue."""
  for testcase_1_id, testcase_1 in six.iteritems(testcase_map):
    for testcase_2_id, testcase_2 in six.iteritems(testcase_map):
      # Rule: Don't group the same testcase and use different combinations for
      # comparisons.
      if testcase_1_id <= testcase_2_id:
        continue

      # Rule: If both testcase have the same group id, then no work to do.
      if testcase_1.group_id == testcase_2.group_id and testcase_1.group_id:
        continue

      # Rule: Check both testcase have an associated issue id.
      if testcase_1.issue_id is None or testcase_2.issue_id is None:
        continue

      # Rule: Check both testcase are under the same project.
      if testcase_1.project_name != testcase_2.project_name:
        continue

      # Rule: Group testcase with same underlying issue.
      if testcase_1.issue_id != testcase_2.issue_id:
        continue

      combine_testcases_into_group(testcase_1, testcase_2, testcase_map)


def _group_testcases_with_similar_states(testcase_map):
  """Group testcases with similar looking crash states."""
  for testcase_1_id, testcase_1 in six.iteritems(testcase_map):
    for testcase_2_id, testcase_2 in six.iteritems(testcase_map):
      # Rule: Don't group the same testcase and use different combinations for
      # comparisons.
      if testcase_1_id <= testcase_2_id:
        continue

      # If both testcase have the same group id, then no work to do.
      if testcase_1.group_id == testcase_2.group_id and testcase_1.group_id:
        continue

      # Rule: Check both testcase are under the same project.
      if testcase_1.project_name != testcase_2.project_name:
        continue

      # Rule: Security bugs should never be grouped with functional bugs.
      if testcase_1.security_flag != testcase_2.security_flag:
        continue

      # Rule: Follow different comparison rules when crash types is one of the
      # ones that have unique crash state (custom ones specifically).
      if (testcase_1.crash_type in data_types.CRASH_TYPES_WITH_UNIQUE_STATE or
          testcase_2.crash_type in data_types.CRASH_TYPES_WITH_UNIQUE_STATE):

        # For grouping, make sure that both crash type and state match.
        if (testcase_1.crash_type != testcase_2.crash_type or
            testcase_1.crash_state != testcase_2.crash_state):
          continue

      else:
        # Rule: For functional bugs, compare for similar crash states.
        if not testcase_1.security_flag:
          crash_comparer = CrashComparer(testcase_1.crash_type,
                                         testcase_2.crash_type)
          if not crash_comparer.is_similar():
            continue

        # Rule: Check for crash state similarity.
        crash_comparer = CrashComparer(testcase_1.crash_state,
                                       testcase_2.crash_state)
        if not crash_comparer.is_similar():
          continue

      combine_testcases_into_group(testcase_1, testcase_2, testcase_map)


def _has_testcase_with_same_params(testcase, testcase_map):
  """Return a bool whether there is another testcase with same params."""
  for other_testcase_id in testcase_map:
    # yapf: disable
    if (testcase.project_name ==
        testcase_map[other_testcase_id].project_name and
        testcase.crash_state ==
        testcase_map[other_testcase_id].crash_state and
        testcase.crash_type ==
        testcase_map[other_testcase_id].crash_type and
        testcase.security_flag ==
        testcase_map[other_testcase_id].security_flag and
        testcase.one_time_crasher_flag ==
        testcase_map[other_testcase_id].one_time_crasher_flag):
      return True
    # yapf: enable

  return False


def _shrink_large_groups_if_needed(testcase_map):
  """Shrinks groups that exceed a particular limit."""

  def _key_func(testcase):
    weight = 0
    if not testcase.one_time_crasher_flag:
      weight |= 2**1
    if testcase.issue_id:
      weight |= 2**2
    return weight

  group_id_with_testcases_map = {}
  for testcase in six.itervalues(testcase_map):
    if not testcase.group_id:
      continue

    if testcase.group_id not in group_id_with_testcases_map:
      group_id_with_testcases_map[testcase.group_id] = [testcase]
    else:
      group_id_with_testcases_map[testcase.group_id].append(testcase)

  for group_id in group_id_with_testcases_map:
    testcases_in_group = group_id_with_testcases_map[group_id]
    if len(testcases_in_group) <= GROUP_MAX_TESTCASE_LIMIT:
      continue

    testcases_in_group = sorted(testcases_in_group, key=_key_func)
    for testcase in testcases_in_group[:-GROUP_MAX_TESTCASE_LIMIT]:
      try:
        testcase_entity = data_handler.get_testcase_by_id(testcase.id)
      except errors.InvalidTestcaseError:
        # Already deleted.
        continue

      if testcase_entity.bug_information:
        continue

      logs.log_warn(('Deleting testcase {testcase_id} due to overflowing group '
                     '{group_id}.').format(
                         testcase_id=testcase.id, group_id=testcase.group_id))
      testcase_entity.key.delete()


def group_testcases():
  """Group testcases based on rules like same bug numbers, similar crash
  states, etc."""
  testcase_map = {}
  cached_issue_map = {}

  for testcase_id in data_handler.get_open_testcase_id_iterator():
    try:
      testcase = data_handler.get_testcase_by_id(testcase_id)
    except errors.InvalidTestcaseError:
      # Already deleted.
      continue

    # Remove duplicates early on to avoid large groups.
    if (not testcase.bug_information and not testcase.uploader_email and
        _has_testcase_with_same_params(testcase, testcase_map)):
      logs.log('Deleting duplicate testcase %d.' % testcase_id)
      testcase.key.delete()
      continue

    # Wait for minimization to finish as this might change crash params such
    # as type and may mark it as duplicate / closed.
    if not testcase.minimized_keys:
      continue

    # Store needed testcase attributes into |testcase_map|.
    testcase_map[testcase_id] = TestcaseAttributes(testcase_id)
    testcase_attributes = testcase_map[testcase_id]
    for attribute_name in FORWARDED_ATTRIBUTES:
      setattr(testcase_attributes, attribute_name,
              getattr(testcase, attribute_name))

    # Store original issue mappings in the testcase attributes.
    if testcase.bug_information:
      issue_id = int(testcase.bug_information)
      project_name = testcase.project_name

      if (project_name in cached_issue_map and
          issue_id in cached_issue_map[project_name]):
        testcase_attributes.issue_id = (
            cached_issue_map[project_name][issue_id])
      else:
        issue_tracker = issue_tracker_utils.get_issue_tracker_for_testcase(
            testcase)
        if not issue_tracker:
          logs.log_error(
              'Unable to access issue tracker for issue %d.' % issue_id)
          testcase_attributes.issue_id = issue_id
          continue

        # Determine the original issue id traversing the list of duplicates.
        try:
          issue = issue_tracker.get_original_issue(issue_id)
          original_issue_id = int(issue.id)
        except:
          # If we are unable to access the issue, then we can't determine
          # the original issue id. Assume that it is the same as issue id.
          logs.log_error(
              'Unable to determine original issue for issue %d.' % issue_id)
          testcase_attributes.issue_id = issue_id
          continue

        if project_name not in cached_issue_map:
          cached_issue_map[project_name] = {}
        cached_issue_map[project_name][issue_id] = original_issue_id
        cached_issue_map[project_name][original_issue_id] = original_issue_id
        testcase_attributes.issue_id = original_issue_id

  # No longer needed. Free up some memory.
  cached_issue_map.clear()

  _group_testcases_with_similar_states(testcase_map)
  _group_testcases_with_same_issues(testcase_map)
  _shrink_large_groups_if_needed(testcase_map)
  group_leader.choose(testcase_map)

  # TODO(aarya): Replace with an optimized implementation using dirty flag.
  # Update the group mapping in testcase object.
  for testcase_id in data_handler.get_open_testcase_id_iterator():
    if testcase_id not in testcase_map:
      # A new testcase that was just created. Skip for now, will be grouped in
      # next iteration of group task.
      continue

    # If we are part of a group, then calculate the number of testcases in that
    # group and lowest issue id of issues associated with testcases in that
    # group.
    updated_group_id = testcase_map[testcase_id].group_id
    updated_is_leader = testcase_map[testcase_id].is_leader
    updated_group_id_count = 0
    updated_group_bug_information = 0
    if updated_group_id:
      for other_testcase in six.itervalues(testcase_map):
        if other_testcase.group_id != updated_group_id:
          continue
        updated_group_id_count += 1

        # Update group issue id to be lowest issue id in the entire group.
        if other_testcase.issue_id is None:
          continue
        if (not updated_group_bug_information or
            updated_group_bug_information > other_testcase.issue_id):
          updated_group_bug_information = other_testcase.issue_id

    # If this group id is used by only one testcase, then remove it.
    if updated_group_id_count == 1:
      data_handler.delete_group(updated_group_id, update_testcases=False)
      updated_group_id = 0
      updated_group_bug_information = 0
      updated_is_leader = True

    try:
      testcase = data_handler.get_testcase_by_id(testcase_id)
    except errors.InvalidTestcaseError:
      # Already deleted.
      continue

    is_changed = (
        (testcase.group_id != updated_group_id) or
        (testcase.group_bug_information != updated_group_bug_information) or
        (testcase.is_leader != updated_is_leader))

    if not is_changed:
      # If nothing is changed, no more work to do. It's faster this way.
      continue

    testcase.group_bug_information = updated_group_bug_information
    testcase.group_id = updated_group_id
    testcase.is_leader = updated_is_leader
    testcase.put()
    logs.log(
        'Updated testcase %d group to %d.' % (testcase_id, updated_group_id))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Uploads test cases from blackbox fuzzers for coverage collection."""

import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import builtin_fuzzers
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.google_cloud_utils import gsutil
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

LIST_FILE_BASENAME = 'file_list.txt'
TESTCASES_PER_DAY = 1000


def upload_testcases_if_needed(fuzzer_name, testcase_list, testcase_directory,
                               data_directory):
  """Upload test cases from the list to a cloud storage bucket."""
  # Since builtin fuzzers have a coverage minimized corpus, no need to upload
  # test case samples for them.
  if fuzzer_name in builtin_fuzzers.BUILTIN_FUZZERS:
    return

  bucket_name = local_config.ProjectConfig().get(
      'coverage.fuzzer-testcases.bucket')
  if not bucket_name:
    return

  files_list = []
  has_testcases_in_testcase_directory = False
  has_testcases_in_data_directory = False
  for testcase_path in testcase_list:
    if testcase_path.startswith(testcase_directory):
      files_list.append(os.path.relpath(testcase_path, testcase_directory))
      has_testcases_in_testcase_directory = True
    elif testcase_path.startswith(data_directory):
      files_list.append(os.path.relpath(testcase_path, data_directory))
      has_testcases_in_data_directory = True
  if not files_list:
    return

  formatted_date = str(utils.utcnow().date())
  gcs_base_url = 'gs://{bucket_name}/{date}/{fuzzer_name}/'.format(
      bucket_name=bucket_name, date=formatted_date, fuzzer_name=fuzzer_name)

  runner = gsutil.GSUtilRunner()
  batch_directory_blobs = storage.list_blobs(gcs_base_url)
  total_testcases = 0
  for blob in batch_directory_blobs:
    if not blob.endswith(LIST_FILE_BASENAME):
      continue

    list_gcs_url = storage.get_cloud_storage_file_path(bucket_name, blob)
    data = storage.read_data(list_gcs_url)
    if not data:
      logs.log_error('Read no data from test case list at {gcs_url}'.format(
          gcs_url=list_gcs_url))
      continue

    total_testcases += len(data.splitlines())

    # If we've already uploaded enough test cases for this fuzzer today, return.
    if total_testcases >= TESTCASES_PER_DAY:
      return

  # Cap the number of files.
  testcases_limit = min(len(files_list), TESTCASES_PER_DAY - total_testcases)
  files_list = files_list[:testcases_limit]

  # Upload each batch of tests to its own unique sub-bucket.
  identifier = environment.get_value('BOT_NAME') + str(utils.utcnow())
  gcs_base_url += utils.string_hash(identifier)

  list_gcs_url = gcs_base_url + '/' + LIST_FILE_BASENAME
  if not storage.write_data('\n'.join(files_list).encode('utf-8'),
                            list_gcs_url):
    return

  if has_testcases_in_testcase_directory:
    # Sync everything in |testcase_directory| since it is fuzzer-generated.
    runner.rsync(testcase_directory, gcs_base_url)

  if has_testcases_in_data_directory:
    # Sync all fuzzer generated testcase in data bundle directory.
    runner.rsync(
        data_directory,
        gcs_base_url,
        exclusion_pattern=('(?!.*{fuzz_prefix})'.format(
            fuzz_prefix=testcase_manager.FUZZ_PREFIX)))

    # Sync all possible resource dependencies as a best effort. It matches
    # |resources-| prefix that a fuzzer can use to indicate resources. Also, it
    # matches resources directory that Chromium web_tests use for dependencies.
    runner.rsync(
        data_directory, gcs_base_url, exclusion_pattern='(?!.*resource)')

  logs.log('Synced {count} test cases to {gcs_url}.'.format(
      count=len(files_list), gcs_url=gcs_base_url))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for app specific trials and experiments."""

import unittest

from clusterfuzz._internal.bot.tasks import trials
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class TrialsTest(unittest.TestCase):
  """Tests for trials."""

  def setUp(self):
    test_helpers.patch_environ(self, env={'APP_ARGS': '-x'})

    data_types.Trial(app_name='app_1', probability=0.5, app_args='--a1').put()

    data_types.Trial(app_name='app_2', probability=0.4, app_args='--b1').put()
    data_types.Trial(app_name='app_2', probability=0.2, app_args='--b2').put()

    data_types.Trial(app_name='app_3', probability=1.0, app_args='--c1').put()
    data_types.Trial(app_name='app_3', probability=0.2, app_args='--c2').put()
    data_types.Trial(app_name='app_3', probability=0.2, app_args='--c3').put()

    test_helpers.patch(self, ['random.random'])

  def test_no_effect_on_no_match(self):
    """Ensure that no additional flags are added if a binary has no trials."""
    self.mock.random.return_value = 0.0
    environment.set_value('APP_NAME', 'app_0')
    trial_selector = trials.Trials()
    trial_selector.setup_additional_args_for_app()
    self.assertEqual(environment.get_value('APP_ARGS'), '-x')
    self.assertIsNone(environment.get_value('TRIAL_APP_ARGS'))

  def test_trial_selected_one_option(self):
    """Ensure that the expected flags are added if a trial is selected."""
    self.mock.random.return_value = 0.3
    environment.set_value('APP_NAME', 'app_1')
    trial_selector = trials.Trials()
    trial_selector.setup_additional_args_for_app()
    self.assertEqual(environment.get_value('APP_ARGS'), '-x --a1')
    self.assertEqual(environment.get_value('TRIAL_APP_ARGS'), '--a1')

  def test_trial_not_selected(self):
    """Ensure no additional flags if a trial was not selected."""
    self.mock.random.return_value = 0.5
    environment.set_value('APP_NAME', 'app_2')
    trial_selector = trials.Trials()
    trial_selector.setup_additional_args_for_app()
    self.assertEqual(environment.get_value('APP_ARGS'), '-x')
    self.assertIsNone(environment.get_value('TRIAL_APP_ARGS'))

  def test_multiple_trial_selection(self):
    """Ensure that we can suggest the second trial in a batch of multiple."""
    self.mock.random.return_value = 0.1
    environment.set_value('APP_NAME', 'app_3')
    trial_selector = trials.Trials()
    trial_selector.setup_additional_args_for_app()
    self.assertEqual(environment.get_value('APP_ARGS'), '-x --c1 --c2 --c3')
    self.assertEqual(environment.get_value('TRIAL_APP_ARGS'), '--c1 --c2 --c3')

  def test_selection_for_windows_executable(self):
    """Ensure that flags are added when the app name ends in ".exe"."""
    self.mock.random.return_value = 0.3
    environment.set_value('APP_NAME', 'app_1.exe')
    trial_selector = trials.Trials()
    trial_selector.setup_additional_args_for_app()
    self.assertEqual(environment.get_value('APP_ARGS'), '-x --a1')
    self.assertEqual(environment.get_value('TRIAL_APP_ARGS'), '--a1')

  def test_selection_for_android_apk(self):
    """Ensure that flags are added for the Android APK format."""
    self.mock.random.return_value = 0.3
    environment.set_value('APP_NAME', 'App_1.apk')
    trial_selector = trials.Trials()
    trial_selector.setup_additional_args_for_app()
    self.assertEqual(environment.get_value('APP_ARGS'), '-x --a1')
    self.assertEqual(environment.get_value('TRIAL_APP_ARGS'), '--a1')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""libClusterfuzz package."""

import os

if not os.getenv('ROOT_DIR') and not os.getenv('GAE_ENV'):
  # If ROOT_DIR isn't set by the time we import this and we're not on GAE,
  # assume we're libClusterFuzz.
  this_dir = os.path.dirname(os.path.abspath(__file__))
  os.environ['CONFIG_DIR_OVERRIDE'] = os.path.join(this_dir, 'lib-config')
  os.environ['ROOT_DIR'] = os.path.dirname(os.path.dirname(this_dir))
  os.environ['LIB_CF'] = 'True'

  from ._internal.system import environment
  environment.set_default_vars()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""syzkaller fuzzer."""
import copy
import fnmatch
import os
import re
import tempfile
import threading
import time

import requests

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.syzkaller import config
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.android import kernel_utils
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz.fuzz import engine

LOCAL_HOST = '127.0.0.1'
RAWCOVER_RETRIEVE_INTERVAL = 180  # retrieve rawcover every 180 seconds
REPRODUCE_REGEX = re.compile(r'reproduced (\d+) crashes')


def get_work_dir():
  """Return work directory for Syzkaller."""
  work_dir = os.path.join(
      environment.get_value('FUZZ_INPUTS_DISK'), 'syzkaller')

  os.makedirs(work_dir, exist_ok=True)
  return work_dir


def get_config():
  """Get arguments for a given fuzz target."""
  device_serial = environment.get_value('ANDROID_SERIAL')
  build_dir = environment.get_value('BUILD_DIR')
  temp_dir = fuzzer_utils.get_temp_dir()

  binary_path = os.path.join(build_dir, 'syzkaller')
  json_config_path = os.path.join(temp_dir, 'config.json')
  default_vmlinux_path = os.path.join('/tmp', device_serial, 'vmlinux')
  vmlinux_path = environment.get_value('VMLINUX_PATH', default_vmlinux_path)

  syzhub_address = environment.get_value('SYZHUB_ADDRESS')
  syzhub_client = environment.get_value('SYZHUB_CLIENT')
  syzhub_key = environment.get_value('SYZHUB_KEY')
  on_cuttlefish = environment.is_android_cuttlefish()

  config.generate(
      serial=device_serial,
      work_dir_path=get_work_dir(),
      binary_path=binary_path,
      vmlinux_path=vmlinux_path,
      config_path=json_config_path,
      kcov=True,
      reproduce=False,
      syzhub_address=syzhub_address,
      syzhub_client=syzhub_client,
      syzhub_key=syzhub_key,
      on_cuttlefish=on_cuttlefish)
  return ['-config', json_config_path]


def get_cover_file_path():
  """Return location of coverage file for Syzkaller."""
  return os.path.join(get_work_dir(), 'coverfile')


def get_runner(fuzzer_path):
  """Return a syzkaller runner object."""
  return AndroidSyzkallerRunner(fuzzer_path)


def _upload_kernel_coverage_data(kcov_path, kernel_bid):
  """Upload kcov data to a cloud storage bucket."""
  bucket_name = local_config.ProjectConfig().get('coverage.reports.bucket')
  if not bucket_name:
    return

  formatted_date = str(utils.utcnow().date().isoformat())
  identifier = environment.get_value('BOT_NAME') + str(
      utils.utcnow().isoformat())

  gcs_url = (f'gs://{bucket_name}/syzkaller/{formatted_date}/{kernel_bid}/'
             f'{identifier}')
  if storage.copy_file_to(kcov_path, gcs_url):
    logs.log(f'Copied kcov data to {gcs_url}.')


class LoopingTimer(threading.Timer):
  """Extend Timer to loop every interval seconds."""

  def __init__(self, interval, function, args=None, kwargs=None):
    super(LoopingTimer, self).__init__(
        interval, function, args=args, kwargs=kwargs)

  def run(self):
    # loops until self.cancel()
    while not self.finished.is_set():
      self.finished.wait(self.interval)
      self.function(*self.args, **self.kwargs)


class AndroidSyzkallerRunner(new_process.UnicodeProcessRunner):
  """Syzkaller runner."""

  def __init__(self, executable_path):
    """Inits the AndroidSyzkallerRunner.

    Args:
      executable_path: Path to the fuzzer executable.
    """
    super().__init__(executable_path=executable_path)
    self._port = None

  def get_command(self, additional_args=None):
    """Process.get_command override."""
    base_command = super().get_command(additional_args=additional_args)

    return base_command

  def get_port(self, pid: int) -> int or None:
    """Find localhost port where syzkaller is connected."""

    if self._port is not None:
      return self._port

    import psutil  # pylint: disable=g-import-not-at-top

    for connection in psutil.net_connections():
      if connection.pid != pid:
        continue

      local_address = connection.laddr
      if (local_address.ip == LOCAL_HOST and
          connection.status == psutil.CONN_LISTEN):
        self._port = local_address.port
        logs.log(f'Syzkaller listening at: http://localhost:{self._port}')
        return self._port

    # No connection found
    return None

  def _create_empty_testcase_file(self):
    """Create an empty testcase file in temporary directory."""
    _, path = tempfile.mkstemp(dir=fuzzer_utils.get_temp_dir())
    return path

  def _crash_was_reproducible(self, output):
    reproducible = False
    if 'all done.' in output:
      search = REPRODUCE_REGEX.search(output)
      if search and search.group(1) and search.group(1) > '0':
        reproducible = True
    return int(reproducible)

  def repro(self, repro_timeout, repro_args):
    """This is where crash repro'ing is done.
    Args:
      repro_timeout: The maximum time in seconds that repro job is allowed
          to run for.
      repro_args: A sequence of arguments to be passed to the executable.
    """
    logs.log('Running Syzkaller testcase.')
    additional_args = copy.copy(repro_args)
    result = self.run_and_wait(additional_args, timeout=repro_timeout)
    result.return_code = self._crash_was_reproducible(result.output)

    if result.return_code:
      logs.log('Successfully reproduced crash.')
    else:
      logs.log('Failed to reproduce crash.')
    logs.log('Syzkaller repro testcase stopped.')
    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def save_rawcover_output(self, pid: int):
    """Find syzkaller port and write rawcover data to a file."""

    port = self.get_port(pid)
    if port is None:
      logs.log_warn('Could not find Syzkaller port')
      return

    try:
      rawcover = requests.get(f'http://localhost:{port}/rawcover').text
    except requests.exceptions.ConnectionError:
      logs.log_warn('Connection to Syzkaller Failed')
      return

    if not rawcover or rawcover.startswith('coverage is not ready'):
      logs.log_warn('Syzkaller rawcover not yet loaded')
      return

    file_path = get_cover_file_path()
    with open(file_path, 'w+') as f:
      f.write(rawcover)
      logs.log(f'Writing syzkaller rawcover to {file_path}')

  def run_and_loop(self, *args, timeout=None,
                   **kwargs) -> new_process.ProcessResult:
    """Adds looping call to run_and_wait method.

    This method adds LoopingTimer() that continuously executes a function
    that gets / saves rawcover data from Syzkaller.

    Args:
      *args: args for self.run()
      timeout: timeout in seconds to stop Syzkaller
      **kwargs: kwargs for self.run()
    Returns:
      new_process.ProcessResult from Syzkaller
    """
    process = self.run(*args, **kwargs)
    pid = process.popen.pid
    logs.log(f'Syzkaller pid = {pid}')

    looping_timer = LoopingTimer(
        RAWCOVER_RETRIEVE_INTERVAL,
        self.save_rawcover_output,
        args=[pid],
    )
    looping_timer.start()

    try:
      if not timeout:
        start_time = time.time()
        output = process.communicate()[0]
        return new_process.ProcessResult(process.command, process.poll(),
                                         output,
                                         time.time() - start_time, False)

      result = new_process.wait_process(
          process,
          timeout=timeout,
          input_data=None,
          terminate_before_kill=False,
          terminate_wait_time=None,
      )
      result.command = process.command
      result.output = str(result.output)

      return result
    finally:
      looping_timer.cancel()

  def fuzz(
      self,
      fuzz_timeout,
      additional_args,
      unused_additional_args=None,
      unused_extra_env=None,
  ) -> engine.FuzzResult:
    """This is where actual syzkaller fuzzing is done.

    Args:
      fuzz_timeout (float): The maximum time in seconds that fuzz job is allowed
          to run for.
      additional_args: A sequence of additional arguments to be passed to
          the executable.
    Returns:
      engine.FuzzResult
    """

    def _filter_log(content):
      """Filter unneeded content from log."""
      result = ''
      strip_regex = re.compile(r'^c\d+\s+\d+\s')
      for line in content.splitlines():
        result += strip_regex.sub('', line) + '\n'
      return result

    logs.log('Running Syzkaller.')
    additional_args = copy.copy(additional_args)

    # Save kernel_bid for later in case the device is down.
    _, kernel_bid = kernel_utils.get_kernel_hash_and_build_id()

    fuzz_result = self.run_and_loop(additional_args, timeout=fuzz_timeout)
    logs.log('Syzkaller stopped, fuzzing timed out: {}'.format(
        fuzz_result.time_executed))

    fuzz_logs = (fuzz_result.output or '') + '\n'
    crashes = []
    parsed_stats = {}
    visited = set()
    for subdir, _, files in os.walk(get_work_dir()):
      for file in files:
        # Each crash typically have 2 files: reportN and logN. Similar crashes
        # are grouped together in subfolders. unique_crash puts together the
        # subfolder name and reportN.
        unique_crash = os.path.join(subdir, file)
        if fnmatch.fnmatch(file, 'report*') and unique_crash not in visited:
          visited.add(unique_crash)
          log_content = _filter_log(
              utils.read_data_from_file(
                  os.path.join(subdir, file), eval_data=False).decode('utf-8'))
          fuzz_logs += log_content + '\n'

          # Since each crash (report file) has a corresponding log file
          # that contains the syscalls that caused the crash. This file is
          # located in the same subfolder and has the same number.
          # E.g. ./439c37d288d4f26a33a6c7e5c57a97791453a447/report15 and
          # ./439c37d288d4f26a33a6c7e5c57a97791453a447/log15.
          crash_testcase_file_path = os.path.join(subdir,
                                                  'log' + file[len('report'):])

          # TODO(hzawawy): Parse stats information and add them to FuzzResult.

          if crash_testcase_file_path:
            reproduce_arguments = [unique_crash]
            actual_duration = int(fuzz_result.time_executed)
            # Write the new testcase.
            # Copy crash testcase contents into the main testcase path.
            crashes.append(
                engine.Crash(crash_testcase_file_path, log_content,
                             reproduce_arguments, actual_duration))

    _upload_kernel_coverage_data(get_cover_file_path(), kernel_bid)
    return engine.FuzzResult(fuzz_logs, fuzz_result.command, crashes,
                             parsed_stats, fuzz_result.time_executed)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""archive tests."""
import os
import tempfile
import unittest

from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers

TESTDATA_PATH = os.path.join(os.path.dirname(__file__), 'archive_data')


class UnpackTest(unittest.TestCase):
  """Unpack tests."""

  def test_unpack_file_with_cwd_prefix(self):
    """Test unpack with trusted=False passes with file having './' prefix."""
    tgz_path = os.path.join(TESTDATA_PATH, 'cwd-prefix.tgz')
    output_directory = tempfile.mkdtemp(prefix='cwd-prefix')
    archive.unpack(tgz_path, output_directory, trusted=False)

    test_file_path = os.path.join(output_directory, 'test')
    self.assertTrue(os.path.exists(test_file_path))
    self.assertEqual(open(test_file_path).read(), 'abc\n')

    shell.remove_directory(output_directory)


class IteratorTest(unittest.TestCase):
  """Tests for the archive.iterator function."""

  def test_tar_xz(self):
    """Test that a .tar.xz file is handled properly by iterator()."""
    tar_xz_path = os.path.join(TESTDATA_PATH, 'archive.tar.xz')
    expected_results = {'archive_dir/hi': b'hi\n', 'archive_dir/bye': b'bye\n'}
    actual_results = {
        archive_file.name: archive_file.handle.read()
        for archive_file in archive.iterator(tar_xz_path)
        if archive_file.handle
    }
    self.assertEqual(actual_results, expected_results)

  def test_cwd_prefix(self):
    """Test that a .tgz file with cwd prefix is handled."""
    tgz_path = os.path.join(TESTDATA_PATH, 'cwd-prefix.tgz')
    expected_results = {'./test': b'abc\n'}
    actual_results = {
        archive_file.name: archive_file.handle.read()
        for archive_file in archive.iterator(tgz_path)
        if archive_file.handle
    }
    self.assertEqual(actual_results, expected_results)

  def test_tar_xz_broken_links(self):
    """Test that a .tar file with broken links is handled properly by
    iterator()."""
    helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_warn'])

    archive_name = 'broken-links.tar.xz'
    archive_path = os.path.join(TESTDATA_PATH, archive_name)

    # Get the results we expect from iterator().
    actual_results = []
    for archive_file in archive.iterator(archive_path):
      if archive_file.handle is not None:
        actual_results.append((archive_file.name, archive_file.size,
                               archive_file.handle.read()))
      else:
        actual_results.append((archive_file.name, archive_file.size, None))

    # Check that iterator returns what we expect it to.
    expected_results = [
        ('testdir', 0, None),
        ('testdir/1', 0, None),
        ('testdir/1/a', 12, b'hello world\n'),
        ('testdir/2', 0, None),
        ('testdir/2/c', 0, b'hello world\n'),  # Working link
        ('testdir/2/a', 0, None),
        ('testdir/2/b', 0, None)
    ]

    self.assertEqual(expected_results, actual_results)

    # Check that iterator calls log_warn on a broken link.
    self.mock.log_warn.assert_called_with(
        'Check archive %s for broken links.' % archive_path,
        error_filepaths=['testdir/2/a', 'testdir/2/b'])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for sanitizer functions."""

import os

from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.platforms.android import constants
from clusterfuzz._internal.platforms.android import sanitizer
from clusterfuzz._internal.platforms.android import settings
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import android_helpers
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'sanitizer_data')


class GetOptionsFilePathTest(android_helpers.AndroidTest):
  """Tests get_options_file_path."""

  def test_regular_build(self):
    """Test that options file path is returned inside device temp dir when
    device has ASan setup with partial instrumentation using asan_device_setup
    script."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.platforms.android.settings.get_sanitizer_tool_name'
    ])
    self.mock.get_sanitizer_tool_name.return_value = None

    self.assertEqual('/data/local/tmp/asan.options',
                     sanitizer.get_options_file_path('asan'))

  def test_system_asan_build(self):
    """Test that options file path is returned inside /system when device is
    setup with a full-system ASan build."""
    test_helpers.patch(self, [
        'clusterfuzz._internal.platforms.android.settings.get_sanitizer_tool_name'
    ])
    self.mock.get_sanitizer_tool_name.return_value = 'asan'

    self.assertEqual('/system/asan.options',
                     sanitizer.get_options_file_path('asan'))

  def test_invalid(self):
    """Test that no options file path is returned with an invalid sanitizer
    name."""
    self.assertEqual(None, sanitizer.get_options_file_path('invalid'))

  def test_unsupported(self):
    """Test that no options file path is returned with an unsupported sanitizer
    e.g. UBSan, MSan."""
    self.assertEqual(None, sanitizer.get_options_file_path('msan'))
    self.assertEqual(None, sanitizer.get_options_file_path('tsan'))
    self.assertEqual(None, sanitizer.get_options_file_path('ubsan'))


class SetOptionsTest(android_helpers.AndroidTest):
  """Tests set_options."""

  def setUp(self):
    """Setup for set options test."""
    super(SetOptionsTest, self).setUp()

    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_error'])

    if settings.get_sanitizer_tool_name():
      self.skipTest('This test is not applicable on a system sanitizer build.')

    # Clear and create temporary directory on device.
    self.device_temp_dir = constants.DEVICE_TMP_DIR
    adb.remove_directory(self.device_temp_dir, recreate=True)

  def test(self):
    """Test that options are successfully set with ASan."""
    sanitizer.set_options('ASAN', 'a=b:c=d')
    self.assertEqual('a=b:c=d',
                     adb.read_data_from_file('/data/local/tmp/asan.options'))
    self.assertEqual(0, self.mock.log_error.call_count)

  def test_unsupported(self):
    """Test that options are not set with an unsupported sanitizer e.g.
    UBSan, MSan, etc."""
    sanitizer.set_options('UBSAN', 'a=b:c=d')
    self.assertFalse(adb.file_exists('/data/local/tmp/ubsan.options'))
    self.assertEqual(1, self.mock.log_error.call_count)

  def test_invalid(self):
    """Test that options are not set with an invalid sanitizer name."""
    sanitizer.set_options('invalid', 'a=b:c=d')
    self.assertEqual(1, self.mock.log_error.call_count)


class SetupASanIfNeededTest(android_helpers.AndroidTest):
  """Tests setup_asan_if_needed."""

  def setUp(self):
    """Setup for setup ASan if needed test."""
    super(SetupASanIfNeededTest, self).setUp()

    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_error'])

    if settings.get_sanitizer_tool_name():
      self.skipTest('This test is not applicable on a system sanitizer build.')

    environment.set_value('ASAN_DEVICE_SETUP', True)
    environment.set_value('JOB_NAME', 'android_asan_chrome')
    environment.set_value('APP_DIR', DATA_PATH)

  def test(self):
    """Test that ASan instrumentation can be successfully set up on device."""
    adb.revert_asan_device_setup_if_needed()
    environment.reset_current_memory_tool_options()
    sanitizer.setup_asan_if_needed()
    self.assertEqual(0, self.mock.log_error.call_count)
    self.assertTrue(adb.file_exists('/system/bin/asanwrapper'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for remote_process."""

import os

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.untrusted_runner import config
from clusterfuzz._internal.bot.untrusted_runner import file_host
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class FileHostTest(fake_filesystem_unittest.TestCase):
  """FileHost tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.untrusted_runner.host.stub',
    ])

    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)

  def test_create_directory(self):
    """Test file_host.create_directory."""
    result = untrusted_runner_pb2.CreateDirectoryResponse(result=True)
    self.mock.stub().CreateDirectory.return_value = result
    self.assertTrue(file_host.create_directory('/path', True))

    result = untrusted_runner_pb2.CreateDirectoryResponse(result=False)
    self.mock.stub().CreateDirectory.return_value = result
    self.assertFalse(file_host.create_directory('/path', True))

  def test_remove_directory(self):
    """Test file_host.remove_directory."""
    result = untrusted_runner_pb2.RemoveDirectoryResponse(result=True)
    self.mock.stub().RemoveDirectory.return_value = result
    self.assertTrue(file_host.remove_directory('/path', True))

    result = untrusted_runner_pb2.RemoveDirectoryResponse(result=False)
    self.mock.stub().RemoveDirectory.return_value = result
    self.assertFalse(file_host.remove_directory('/path', True))

  def test_copy_file_to_worker(self):
    """Test file_host.copy_file_to_worker."""
    contents = (b'A' * config.FILE_TRANSFER_CHUNK_SIZE +
                b'B' * config.FILE_TRANSFER_CHUNK_SIZE +
                b'C' * config.FILE_TRANSFER_CHUNK_SIZE)

    self.fs.create_file('/file', contents=contents)

    def mock_copy_file_to(iterator, metadata):
      """Mock copy file to."""
      chunks = [chunk.data for chunk in iterator]
      self.assertEqual(3, len(chunks))

      self.assertEqual([('path-bin', b'/file')], metadata)

      data = b''.join(chunks)
      self.assertEqual(data, contents)

      return untrusted_runner_pb2.CopyFileToResponse(result=True)

    self.mock.stub().CopyFileTo.side_effect = mock_copy_file_to
    self.assertTrue(file_host.copy_file_to_worker('/file', '/file'))

  def test_write_data_to_worker(self):
    """Test file_host.write_data_to_worker."""
    contents = (b'A' * config.FILE_TRANSFER_CHUNK_SIZE +
                b'B' * config.FILE_TRANSFER_CHUNK_SIZE +
                b'C' * config.FILE_TRANSFER_CHUNK_SIZE)

    result = untrusted_runner_pb2.CopyFileToResponse(result=True)
    self.mock.stub().CopyFileTo.return_value = result

    self.assertTrue(file_host.write_data_to_worker(contents, '/file'))
    call_args = self.mock.stub().CopyFileTo.call_args
    self.assertEqual(call_args[1], {'metadata': [('path-bin', b'/file')]})

    chunks = [chunk.data for chunk in call_args[0][0]]
    self.assertEqual(len(chunks), 3)

    data = b''.join(chunks)
    self.assertEqual(data, contents)

  def test_copy_file_from_worker(self):
    """Test file_host.copy_file_from_worker."""
    mock_response = mock.MagicMock()
    mock_response.trailing_metadata.return_value = (('result', 'ok'),)
    mock_response.__iter__.return_value = iter([
        untrusted_runner_pb2.FileChunk(data=b'A'),
        untrusted_runner_pb2.FileChunk(data=b'B'),
        untrusted_runner_pb2.FileChunk(data=b'C'),
    ])

    self.mock.stub().CopyFileFrom.return_value = mock_response

    self.assertTrue(file_host.copy_file_from_worker('/file', '/file'))
    with open('/file') as f:
      self.assertEqual(f.read(), 'ABC')

  def test_copy_file_from_worker_failure(self):
    """Test file_host.copy_file_from_worker (failure)."""
    mock_response = mock.MagicMock()
    mock_response.trailing_metadata.return_value = (('result', 'invalid-path'),)
    self.mock.stub().CopyFileFrom.return_value = mock_response

    self.assertFalse(file_host.copy_file_from_worker('/file', '/file'))
    self.assertFalse(os.path.exists('/file'))

  def test_stat(self):
    """Test file_host.stat."""
    result = untrusted_runner_pb2.StatResponse(
        result=True, st_mode=0, st_size=1, st_atime=2, st_mtime=3, st_ctime=4)

    self.mock.stub().Stat.return_value = result
    self.assertEqual(result, file_host.stat('/path'))

  def test_stat_error(self):
    """Test file_host.stat error."""
    result = untrusted_runner_pb2.StatResponse(
        result=False, st_mode=0, st_size=1, st_atime=2, st_mtime=3, st_ctime=4)

    self.mock.stub().Stat.return_value = result
    self.assertIsNone(file_host.stat('/path'))

  @mock.patch(
      'clusterfuzz._internal.bot.untrusted_runner.file_host.remove_directory')
  @mock.patch(
      'clusterfuzz._internal.bot.untrusted_runner.file_host.copy_file_to_worker'
  )
  def test_copy_directory_to_worker(self, mock_copy_file_to_worker,
                                    mock_remove_directory):
    """Test file_host.copy_directory_to_worker."""
    mock_copy_file_to_worker.return_value = True

    self.fs.create_file('/host/dir/file1')
    self.fs.create_file('/host/dir/file2')
    self.fs.create_file('/host/dir/dir2/file3')
    self.fs.create_file('/host/dir/dir2/file4')
    self.fs.create_file('/host/dir/dir2/dir3/file5')

    self.assertTrue(
        file_host.copy_directory_to_worker('/host/dir', '/worker/copied_dir'))
    mock_copy_file_to_worker.assert_has_calls(
        [
            mock.call('/host/dir/file1', '/worker/copied_dir/file1'),
            mock.call('/host/dir/file2', '/worker/copied_dir/file2'),
            mock.call('/host/dir/dir2/file3', '/worker/copied_dir/dir2/file3'),
            mock.call('/host/dir/dir2/file4', '/worker/copied_dir/dir2/file4'),
            mock.call('/host/dir/dir2/dir3/file5',
                      '/worker/copied_dir/dir2/dir3/file5'),
        ],
        any_order=True)

    self.assertTrue(
        file_host.copy_directory_to_worker(
            '/host/dir', '/worker/copied_dir', replace=True))
    mock_copy_file_to_worker.assert_has_calls(
        [
            mock.call('/host/dir/file1', '/worker/copied_dir/file1'),
            mock.call('/host/dir/file2', '/worker/copied_dir/file2'),
            mock.call('/host/dir/dir2/file3', '/worker/copied_dir/dir2/file3'),
            mock.call('/host/dir/dir2/file4', '/worker/copied_dir/dir2/file4'),
            mock.call('/host/dir/dir2/dir3/file5',
                      '/worker/copied_dir/dir2/dir3/file5'),
        ],
        any_order=True)
    mock_remove_directory.assert_called_with(
        '/worker/copied_dir', recreate=True)

    mock_copy_file_to_worker.return_value = False
    self.assertFalse(
        file_host.copy_directory_to_worker('/host/dir', '/worker/copied_dir2'))

  @mock.patch('clusterfuzz._internal.bot.untrusted_runner.file_host.list_files')
  @mock.patch(
      'clusterfuzz._internal.bot.untrusted_runner.file_host.copy_file_from_worker'
  )
  def test_copy_directory_from_worker(self, mock_copy_file_from_worker,
                                      mock_list_files):
    """Test file_host.copy_directory_from_worker."""
    mock_copy_file_from_worker.return_value = True

    mock_list_files.return_value = [
        '/worker/abc',
        '/worker/def',
        '/worker/dir/ghi',
    ]

    self.assertTrue(file_host.copy_directory_from_worker('/worker', '/host'))
    mock_copy_file_from_worker.assert_has_calls(
        [
            mock.call('/worker/abc', '/host/abc'),
            mock.call('/worker/def', '/host/def'),
            mock.call('/worker/dir/ghi', '/host/dir/ghi'),
        ],
        any_order=True)

    mock_list_files.return_value = [
        '/escape',
    ]
    self.assertFalse(file_host.copy_directory_from_worker('/worker', '/host'))

    mock_list_files.return_value = [
        '/worker/../escape',
    ]
    self.assertFalse(file_host.copy_directory_from_worker('/worker', '/host'))

    mock_list_files.return_value = [
        '../escape',
    ]
    self.assertFalse(file_host.copy_directory_from_worker('/worker', '/host'))

  def test_get_cf_worker_path(self):
    """Test get worker path."""
    os.environ['WORKER_ROOT_DIR'] = '/worker'
    local_path = os.path.join(os.environ['ROOT_DIR'], 'a', 'b', 'c')

    self.assertEqual(
        file_host.rebase_to_worker_root(local_path), '/worker/a/b/c')

    local_path = os.environ['ROOT_DIR']
    self.assertEqual(file_host.rebase_to_worker_root(local_path), '/worker')

  def test_get_cf_host_path(self):
    """Test get host path."""
    os.environ['ROOT_DIR'] = '/host'
    os.environ['WORKER_ROOT_DIR'] = '/worker'
    worker_path = os.path.join(os.environ['WORKER_ROOT_DIR'], 'a', 'b', 'c')

    self.assertEqual(file_host.rebase_to_host_root(worker_path), '/host/a/b/c')

    worker_path = os.environ['WORKER_ROOT_DIR']
    self.assertEqual(file_host.rebase_to_host_root(worker_path), '/host')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stacktrace symbolization (host side)."""

from clusterfuzz._internal.protos import untrusted_runner_pb2

from . import host
from . import protobuf_utils


def symbolize_stacktrace(unsymbolized_crash_stacktrace,
                         enable_inline_frames=True):
  """Symbolize stacktrace."""
  request = untrusted_runner_pb2.SymbolizeStacktraceRequest(
      unsymbolized_crash_stacktrace=protobuf_utils.encode_utf8_if_unicode(
          unsymbolized_crash_stacktrace),
      enable_inline_frames=enable_inline_frames)

  response = host.stub().SymbolizeStacktrace(request)
  return response.symbolized_stacktrace



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial datastore setup."""

from google.cloud import monitoring_v3
import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import monitor
from clusterfuzz._internal.metrics import monitoring_metrics

LIBFUZZER_TEMPLATE = """MAX_FUZZ_THREADS = 1
MAX_TESTCASES = 2
FUZZ_TEST_TIMEOUT = 8400
TEST_TIMEOUT = 65
WARMUP_TIMEOUT = 65
BAD_BUILD_CHECK = False
THREAD_ALIVE_CHECK_INTERVAL = 1
REPORT_OOMS_AND_HANGS = True
CORPUS_FUZZER_NAME_OVERRIDE = libFuzzer
ENABLE_GESTURES = False
THREAD_DELAY = 30.0
"""

SYZKALLER_TEMPLATE = """MAX_FUZZ_THREADS = 1
MAX_TESTCASES = 1
FUZZ_TEST_TIMEOUT = 10800
TEST_TIMEOUT = 120
WARMUP_TIMEOUT = 120
BAD_BUILD_CHECK = False
THREAD_ALIVE_CHECK_INTERVAL = 1
REPORT_OOMS_AND_HANGS = True
ENABLE_GESTURES = False
THREAD_DELAY = 30.0
"""

AFL_TEMPLATE = """MAX_FUZZ_THREADS = 1
MAX_TESTCASES = 2
FUZZ_TEST_TIMEOUT = 8400
TEST_TIMEOUT = 30
WARMUP_TIMEOUT = 30
BAD_BUILD_CHECK = False
THREAD_ALIVE_CHECK_INTERVAL = 1
CORPUS_FUZZER_NAME_OVERRIDE = libFuzzer
ADDITIONAL_PROCESSES_TO_KILL = afl-fuzz afl-showmap
ENABLE_GESTURES = False
THREAD_DELAY  = 30.0
"""

HONGGFUZZ_TEMPLATE = """MAX_FUZZ_THREADS = 1
MAX_TESTCASES = 2
FUZZ_TEST_TIMEOUT = 8400
TEST_TIMEOUT = 65
WARMUP_TIMEOUT = 65
BAD_BUILD_CHECK = False
THREAD_ALIVE_CHECK_INTERVAL = 1
CORPUS_FUZZER_NAME_OVERRIDE = libFuzzer
ENABLE_GESTURES = False
THREAD_DELAY = 30.0
"""

GOOGLEFUZZTEST_TEMPLATE = """MAX_FUZZ_THREADS = 1
MAX_TESTCASES = 2
FUZZ_TEST_TIMEOUT = 8400
TEST_TIMEOUT = 65
WARMUP_TIMEOUT = 65
BAD_BUILD_CHECK = False
THREAD_ALIVE_CHECK_INTERVAL = 1
REPORT_OOMS_AND_HANGS = True
ENABLE_GESTURES = False
THREAD_DELAY = 30.0
"""

ENGINE_ASAN_TEMPLATE = ('LSAN = True\n'
                        'ADDITIONAL_ASAN_OPTIONS = '
                        'symbolize=0:'
                        'quarantine_size_mb=64:'
                        'strict_memcmp=1:'
                        'fast_unwind_on_fatal=0:'
                        'allocator_release_to_os_interval_ms=500:'
                        'handle_abort=2:'
                        'handle_segv=2:'
                        'handle_sigbus=2:'
                        'handle_sigfpe=2:'
                        'handle_sigill=2')

ENGINE_MSAN_TEMPLATE = ('ADDITIONAL_MSAN_OPTIONS = '
                        'symbolize=0:'
                        'print_stats=1:'
                        'allocator_release_to_os_interval_ms=500:'
                        'halt_on_error=1:'
                        'handle_abort=2:'
                        'handle_segv=2:'
                        'handle_sigbus=2:'
                        'handle_sigfpe=2:'
                        'handle_sigill=2')

ENGINE_UBSAN_TEMPLATE = ('LSAN = False\n'
                         'ADDITIONAL_UBSAN_OPTIONS = '
                         'symbolize=0:'
                         'allocator_release_to_os_interval_ms=500:'
                         'handle_abort=2:'
                         'handle_segv=2:'
                         'handle_sigbus=2:'
                         'handle_sigfpe=2:'
                         'handle_sigill=2')

PRUNE_TEMPLATE = 'CORPUS_PRUNE = True'

TEMPLATES = {
    'afl': AFL_TEMPLATE,
    'engine_asan': ENGINE_ASAN_TEMPLATE,
    'engine_msan': ENGINE_MSAN_TEMPLATE,
    'engine_ubsan': ENGINE_UBSAN_TEMPLATE,
    'honggfuzz': HONGGFUZZ_TEMPLATE,
    'googlefuzztest': GOOGLEFUZZTEST_TEMPLATE,
    'libfuzzer': LIBFUZZER_TEMPLATE,
    'syzkaller': SYZKALLER_TEMPLATE,
    'prune': PRUNE_TEMPLATE,
}


class BaseBuiltinFuzzerDefaults(object):
  """Default values for a builtin Fuzzer data_type. Note this class should be
  inherited and should not be used directly."""

  def __init__(self):
    # Set defaults for any builtin fuzzer.
    self.revision = 1
    self.file_size = 'builtin'
    self.source = 'builtin'
    self.builtin = True

    # Create attributes that must be set by child classes.
    self.name = None
    self.stats_column_descriptions = None
    self.stats_columns = None
    self.key_id = None

  def create_fuzzer(self):
    """Create a Fuzzer data_type with columns set to the defaults specified by
    this object."""
    assert self.name is not None
    return data_types.Fuzzer(
        id=self.key_id,
        revision=self.revision,
        file_size=self.file_size,
        source=self.source,
        name=self.name,
        builtin=self.builtin,
        stats_column_descriptions=self.stats_column_descriptions,
        stats_columns=self.stats_columns)


class LibFuzzerDefaults(BaseBuiltinFuzzerDefaults):
  """Default values for libFuzzer."""

  def __init__(self):
    super().__init__()
    # Override empty values from parent.
    self.name = 'libFuzzer'
    self.key_id = 1337
    # Use single quotes since the string ends in a double quote.
    # pylint: disable=line-too-long
    self.stats_column_descriptions = '''fuzzer: "Fuzz target"
perf_report: "Link to performance analysis report"
tests_executed: "Number of testcases executed during this time period"
new_crashes: "Number of new unique crashes observed during this time period"
edge_coverage: "Coverage for this fuzz target (number of edges/total)"
cov_report: "Link to coverage report"
corpus_size: "Size of the minimized corpus generated based on code coverage (number of testcases and total size on disk)"
avg_exec_per_sec: "Average number of testcases executed per second"
fuzzing_time_percent: "Percent of expected fuzzing time that is actually spent fuzzing."
new_tests_added: "New testcases added to the corpus during fuzzing based on code coverage"
new_features: "New coverage features based on new tests added to corpus."
regular_crash_percent: "Percent of fuzzing runs that had regular crashes (other than ooms, leaks, timeouts, startup and bad instrumentation crashes)"
oom_percent: "Percent of fuzzing runs that crashed on OOMs (should be 0)"
leak_percent: "Percent of fuzzing runs that crashed on memory leaks (should be 0)"
timeout_percent: "Percent of fuzzing runs that had testcases timeout (should be 0)"
startup_crash_percent: "Percent of fuzzing runs that crashed on startup (should be 0)"
avg_unwanted_log_lines: "Average number of unwanted log lines in fuzzing runs (should be 0)"
total_fuzzing_time_hrs: "Total time in hours for which the fuzzer(s) ran. Will be lower if fuzzer hits a crash frequently."
logs: "Link to fuzzing logs"
corpus_backup: "Backup copy of the minimized corpus generated based on code coverage"'''

    self.stats_columns = """_PERFORMANCE_REPORT as perf_report,
sum(t.number_of_executed_units) as tests_executed,
custom(j.new_crashes) as new_crashes,
_EDGE_COV as edge_coverage,
_COV_REPORT as cov_report,
_CORPUS_SIZE as corpus_size,
avg(t.average_exec_per_sec) as avg_exec_per_sec,
avg(t.fuzzing_time_percent) as fuzzing_time_percent,
sum(t.new_units_added) as new_tests_added,
sum(t.new_features) as new_features,
avg(t.crash_count*100) as regular_crash_percent,
avg(t.oom_count*100) as oom_percent,
avg(t.leak_count*100) as leak_percent,
avg(t.timeout_count*100) as timeout_percent,
avg(t.startup_crash_count*100) as startup_crash_percent,
avg(t.log_lines_unwanted) as avg_unwanted_log_lines,
sum(t.actual_duration/3600.0) as total_fuzzing_time_hrs,
_FUZZER_RUN_LOGS as logs,
_CORPUS_BACKUP as corpus_backup,"""


class AflDefaults(BaseBuiltinFuzzerDefaults):
  """Default values for AFL."""

  def __init__(self):
    super().__init__()
    # Override empty values from parent.
    self.name = 'afl'
    self.key_id = 1338
    # Use single quotes since the string ends in a double quote.
    # pylint: disable=line-too-long
    self.stats_column_descriptions = '''fuzzer: "Fuzz target"
new_crashes: "Number of new unique crashes observed during this time period"
edge_coverage: "Edge coverage for this fuzz target (number of edges / total)"
cov_report: "Link to coverage report"
corpus_size: "Size of the minimized corpus generated based on code coverage (number of testcases and total size on disk)"
avg_exec_per_sec: "Average number of testcases executed per second"
stability: "Percentage of edges that behave deterministically"
new_tests_added: "New testcases added to the corpus during fuzzing based on code coverage"
regular_crash_percent: "Percent of fuzzing runs that had regular crashes (other than startup and bad instrumentation crashes)"
timeout_percent: "Percent of fuzzing runs that had testcases timeout (should be 0)"
startup_crash_percent: "Percent of fuzzing runs that crashed on startup (should be 0)"
avg_unwanted_log_lines: "Average number of unwanted log lines in fuzzing runs (should be 0)"
total_fuzzing_time_hrs: "Total time in hours for which the fuzzer(s) ran. Will be lower if fuzzer hits a crash frequently."
logs: "Link to fuzzing logs"
corpus_backup: "Backup copy of the minimized corpus generated based on code coverage"'''

    self.stats_columns = """custom(j.new_crashes) as new_crashes,
_EDGE_COV as edge_coverage,
_COV_REPORT as cov_report,
_CORPUS_SIZE as corpus_size,
avg(t.average_exec_per_sec) as avg_exec_per_sec,
avg(t.stability) as stability,
sum(t.new_units_added) as new_tests_added,
avg(t.crash_count*100) as regular_crash_percent,
avg(t.timeout_count*100) as timeout_percent,
avg(t.startup_crash_count*100) as startup_crash_percent,
avg(t.log_lines_unwanted) as avg_unwanted_log_lines,
sum(t.actual_duration/3600.0) as total_fuzzing_time_hrs,
_FUZZER_RUN_LOGS as logs,
_CORPUS_BACKUP as corpus_backup,"""


class HonggfuzzDefaults(BaseBuiltinFuzzerDefaults):
  """Default values for honggfuzz."""

  def __init__(self):
    super().__init__()
    self.name = 'honggfuzz'
    self.key_id = 1339


class SyzkallerDefaults(BaseBuiltinFuzzerDefaults):
  """Default values for syzkaller."""

  def __init__(self):
    super().__init__()
    # Override empty values from parent.
    self.name = 'syzkaller'
    self.key_id = 1340


class GoogleFuzzTestDefaults(BaseBuiltinFuzzerDefaults):
  """Default values for googlefuzztest."""

  def __init__(self):
    super().__init__()
    self.name = 'googlefuzztest'
    self.key_id = 1341


def setup_config(non_dry_run):
  """Set up configuration."""
  config = data_types.Config.query().get()
  if not config:
    config = data_types.Config()

    if non_dry_run:
      print('Creating config')
      config.put()
    else:
      print('Skip creating config (dry-run mode)')


def setup_fuzzers(non_dry_run):
  """Set up fuzzers."""
  for fuzzer_defaults in [
      AflDefaults(),
      LibFuzzerDefaults(),
      HonggfuzzDefaults(),
      GoogleFuzzTestDefaults(),
      SyzkallerDefaults()
  ]:
    fuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == fuzzer_defaults.name).get()
    if fuzzer:
      print(fuzzer_defaults.name, 'fuzzer already exists')
      if non_dry_run:
        print('Updating stats metrics.')
        fuzzer.stats_columns = fuzzer_defaults.stats_columns
        fuzzer.stats_column_descriptions = (
            fuzzer_defaults.stats_column_descriptions)
        fuzzer.put()

      continue

    if non_dry_run:
      print('Creating fuzzer', fuzzer_defaults.name)
      fuzzer_defaults.create_fuzzer().put()
    else:
      print('Skip creating fuzzer', fuzzer_defaults.name, '(dry-run mode)')


def setup_templates(non_dry_run):
  """Set up templates."""
  for name, template in six.iteritems(TEMPLATES):
    job = data_types.JobTemplate.query(
        data_types.JobTemplate.name == name).get()
    if job:
      print('Template with name', name, 'already exists.')
      continue

    if non_dry_run:
      print('Creating template', name)
      data_types.JobTemplate(name=name, environment_string=template).put()
    else:
      print('Skip creating template', name, '(dry-run mode)')


def setup_metrics(non_dry_run):
  """Set up metrics."""
  client = monitoring_v3.MetricServiceClient()
  project_name = utils.get_application_id()
  project_path = client.project_path(project_name)

  for name in dir(monitoring_metrics):
    metric = getattr(monitoring_metrics, name)
    if not isinstance(metric, monitor.Metric):
      continue

    descriptor = monitoring_v3.types.MetricDescriptor()
    metric.monitoring_v3_metric_descriptor(descriptor)

    if non_dry_run:
      print('Creating metric', descriptor)
      client.create_metric_descriptor(project_path, descriptor)
    else:
      print('Skip creating metric', descriptor, '(dry-run mode)')


def execute(args):
  """Set up initial Datastore models."""
  setup_config(args.non_dry_run)
  setup_fuzzers(args.non_dry_run)
  setup_templates(args.non_dry_run)

  if not args.local:
    setup_metrics(args.non_dry_run)

  print('Done')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing mutators."""

import random
import struct

DEFAULT_MIN_MUTATIONS = 1
DEFAULT_MAX_MUTATIONS = 20


def get_pack_format_and_mask_for_num_bytes(num_bytes,
                                           signed=False,
                                           little_endian=True):
  """Return the struct pack format and bit mask for the integer values of size
  |num_bytes|."""
  if num_bytes == 1:
    pack_fmt = 'B'
    mask = (1 << 8) - 1
  elif num_bytes == 2:
    pack_fmt = 'H'
    mask = (1 << 16) - 1
  elif num_bytes == 4:
    pack_fmt = 'I'
    mask = (1 << 32) - 1
  elif num_bytes == 8:
    pack_fmt = 'Q'
    mask = (1 << 64) - 1
  else:
    raise ValueError

  if signed:
    pack_fmt = pack_fmt.lower()

  if num_bytes > 1:
    if little_endian:
      pack_fmt = '<' + pack_fmt
    else:
      pack_fmt = '>' + pack_fmt

  return pack_fmt, mask


class MutatorPrimitive(object):
  """A mutator primitive."""

  def __init__(self, ratio=0.0, up_to_ratio=False):
    self.ratio = ratio
    self.up_to_ratio = up_to_ratio

  def mutate_ratio(self):
    """Get the mutation ratio."""
    if self.up_to_ratio:
      return self.ratio * random.random()

    return self.ratio


class BitFlipper(MutatorPrimitive):
  """Flip random bits until the given ratio is satisfied."""

  def __init__(self, ratio=0.0, up_to_ratio=False, contiguous_flips=1):
    super(BitFlipper, self).__init__(ratio, up_to_ratio)
    self.contiguous_flips = contiguous_flips

  def mutate(self, buf):
    """Mutator function."""
    bits_flipped = 0

    num_bits = len(buf) * 8
    ratio = self.mutate_ratio()

    while bits_flipped < int(ratio * len(buf)):
      n = random.randint(0, num_bits - 1)
      for i in range(n, min(num_bits, n + self.contiguous_flips)):
        buf[i // 8] ^= (1 << (i % 8))
        bits_flipped += 1


class BinaryValueAdder(MutatorPrimitive):
  """Add random value to binary values of size |num_bytes| in the buffer until
  the given ratio is satisfied."""

  def __init__(self,
               ratio=0.0,
               up_to_ratio=False,
               num_bytes=1,
               add_range=(-35, 35)):
    super(BinaryValueAdder, self).__init__(ratio, up_to_ratio)
    self.num_bytes = num_bytes
    self.add_range = add_range

    # Assume little endian
    self.pack_fmt, self.mask = get_pack_format_and_mask_for_num_bytes(
        num_bytes, signed=False, little_endian=True)

  def mutate(self, buf):
    """Mutator function."""
    num_choices = len(buf) // self.num_bytes
    changed = 0
    ratio = self.mutate_ratio()

    while changed < int(ratio * num_choices):
      n = random.randint(0, num_choices - 1)
      buf_start = n * self.num_bytes
      buf_end = buf_start + self.num_bytes
      orig = struct.unpack(self.pack_fmt, buf[buf_start:buf_end])[0]

      rand_val = random.randint(self.add_range[0], self.add_range[1])
      new_val = (orig + rand_val) & self.mask
      buf[buf_start:buf_end] = bytearray(struct.pack(self.pack_fmt, new_val))
      changed += 1


class ByteRemover(MutatorPrimitive):
  """Randomly delete |num_bytes| bytes at a time until the ratio is
  satisfied."""

  def __init__(self, ratio=0.0, up_to_ratio=False, num_bytes=1):
    super(ByteRemover, self).__init__(ratio, up_to_ratio)
    self.num_bytes = num_bytes

  def mutate(self, buf):
    """Mutator function."""
    num_choices = len(buf) // self.num_bytes
    changed = 0
    ratio = self.mutate_ratio()

    while changed < int(ratio * len(buf)):
      n = random.randint(0, num_choices - 1)

      del_start = n * self.num_bytes
      del_end = del_start + self.num_bytes

      del buf[del_start:del_end]
      changed += 1


class ByteInserter(MutatorPrimitive):
  """Randomly insert |num_bytes| at a time until the ratio is satisfied."""

  def __init__(self, ratio=0.0, up_to_ratio=False, num_bytes=1):
    super(ByteInserter, self).__init__(ratio, up_to_ratio)
    self.num_bytes = num_bytes

  def mutate(self, buf):
    """Mutator function."""
    num_choices = len(buf) + 1
    inserted = 0
    ratio = self.mutate_ratio()

    while inserted < int(ratio * num_choices):
      # TODO(ochang): context aware
      insert_pos = random.randint(0, num_choices)
      rand_bytes = [random.randint(0, 255) for _ in range(self.num_bytes)]
      buf[insert_pos:insert_pos] = bytearray(rand_bytes)
      inserted += self.num_bytes


class ChunkCopier(MutatorPrimitive):
  """Randomly copy |chunk_ratio * len(buf)| bytes at a time from one location
  to another until the ratio is satisfied."""

  def __init__(self, ratio=0.0, up_to_ratio=False, chunk_ratio=0.1):
    super(ChunkCopier, self).__init__(ratio, up_to_ratio)
    self.chunk_ratio = chunk_ratio

  def mutate(self, buf):
    """Mutator function."""
    changed = 0
    ratio = self.mutate_ratio()
    chunk_size = int(self.chunk_ratio * len(buf))

    while changed < int(len(buf) * ratio):
      copy_from = random.randint(0, len(buf) - chunk_size)
      copy_to = random.randint(0, len(buf) - chunk_size)
      buf[copy_to:copy_to + chunk_size] = buf[copy_from:copy_from + chunk_size]

      changed += chunk_size


class SpecialIntReplacer(MutatorPrimitive):
  """Write special integer values (such as 0, INT_MIN, INT_MAX) of size
  |num_bytes| to random locations in the buffer."""

  def __init__(self, ratio=0.0, up_to_ratio=False, num_bytes=1):
    super(SpecialIntReplacer, self).__init__(ratio, up_to_ratio)
    self.num_bytes = num_bytes

    self.pack_fmt = get_pack_format_and_mask_for_num_bytes(
        num_bytes, signed=False, little_endian=True)[0]

  def mutate(self, buf):
    """Mutator function."""
    num_choices = len(buf) // self.num_bytes
    changed = 0

    # Unsigned representations of signed values
    signed_minimum = 1 << (8 * self.num_bytes - 1)
    signed_maximum = signed_minimum - 1

    # For calculation of values closed to signed minimum/maximum
    max_diff = 1 << (4 * self.num_bytes)

    special_ints = [
        struct.pack(self.pack_fmt, 0),
        # signed minimum
        struct.pack(self.pack_fmt, signed_minimum),
        # value close to signed minimum
        struct.pack(self.pack_fmt,
                    signed_minimum + random.randint(1, max_diff)),
        # signed maximum
        struct.pack(self.pack_fmt, signed_maximum),
        # value close to signed maximum
        struct.pack(self.pack_fmt,
                    signed_maximum - random.randint(1, max_diff)),
        # -1 or unsigned maximum
        struct.pack(self.pack_fmt, (1 << (8 * self.num_bytes)) - 1)
    ]

    ratio = self.mutate_ratio()
    while changed < int(ratio * num_choices):
      n = random.randint(0, num_choices - 1)
      buf_start = n * self.num_bytes
      buf_end = buf_start + self.num_bytes
      buf[buf_start:buf_end] = random.sample(special_ints, 1)[0]
      changed += 1


class SignFlipper(MutatorPrimitive):
  """Flip signs of random integer values of size |num_byte| until the ratio is
  satisfied."""

  def __init__(self, ratio=0.0, up_to_ratio=False, num_bytes=1):
    super(SignFlipper, self).__init__(ratio, up_to_ratio)
    self.num_bytes = num_bytes

    self.pack_fmt, self.mask = get_pack_format_and_mask_for_num_bytes(
        num_bytes, signed=True, little_endian=True)

  def mutate(self, buf):
    """Mutator function."""
    num_choices = len(buf) // self.num_bytes
    changed = 0

    ratio = self.mutate_ratio()
    while changed < int(ratio * num_choices):
      n = random.randint(0, num_choices - 1)
      buf_start = n * self.num_bytes
      buf_end = buf_start + self.num_bytes

      original_value = struct.unpack(self.pack_fmt, buf[buf_start:buf_end])[0]
      buf[buf_start:buf_end] = struct.pack(self.pack_fmt.upper(),
                                           (-original_value) & self.mask)
      changed += 1


class Truncator(MutatorPrimitive):
  """Truncate the buffer to a given ratio."""

  def mutate(self, buf):
    """Mutator function."""
    orig_len = len(buf)
    new_len = int(random.random() * self.mutate_ratio() * orig_len)
    del buf[new_len:]


class CombinedMutator(object):
  """Combination of mutator primitives."""

  def __init__(self, mutators=None, num_mutations_choices=None):
    if num_mutations_choices is None:
      num_mutations_choices = list(
          range(DEFAULT_MIN_MUTATIONS, DEFAULT_MAX_MUTATIONS + 1))
    self.mutators = []
    if mutators is not None:
      for mutator in mutators:
        self.mutators.append({'mutator': mutator[0], 'weight': mutator[1]})
    self.num_mutations_choices = num_mutations_choices

  def add_mutator(self, mutator, weight):
    """Add a mutator."""
    self.mutators.append({'mutator': mutator, 'weight': weight})

  def mutate(self, buf):
    """Mutator function."""
    num_mutations = random.sample(self.num_mutations_choices, 1)[0]
    for _ in range(num_mutations):
      mutator = self.choose_mutator()
      mutator.mutate(buf)

  def choose_mutator(self):
    """Choose a mutator."""
    total_weight = sum([x['weight'] for x in self.mutators])
    n = total_weight * random.random()
    cur = 0.0
    for choice in self.mutators:
      cur += choice['weight']
      if cur > n:
        return choice['mutator']

    return self.mutators[-1]['mutator']



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent the interface for Query. This is important because our access
  control logic needs a unified way to specify conditions for both BigQuery
  query and Datastore query.

  This must be compatible with libs.filters and libs.crash_access."""


class Query(object):
  """Represent the interface for Query."""

  def filter(self, field, value, operator='='):
    """Filter by a single value."""
    raise NotImplementedError

  def filter_in(self, field, values):
    """Filter by multiple values."""
    raise NotImplementedError

  def union(self, *queries):
    """Union all queries with OR conditions."""
    raise NotImplementedError

  def new_subquery(self):
    """Instantiate a query that is compatible with the current query."""
    raise NotImplementedError
